#ifndef __ATM_H
#define __ATM_H


#if !HOST_IS_BIG_ENDIAN
// ATM UNI cell header as seen by an Intel
// processor before swapping the bytes
//
// remember that bitfields are allocated
// from lsb to msb within a byte
//
struct AtmRawHeader
{
    Bit16 VPI_high  : 4;
    Bit16 GFC       : 4;
    Bit16 VCI_high  : 4;
    Bit16 VPI_low   : 4;
    Bit16 VCI_med   : 8;
    Bit16 CLP       : 1;
    Bit16 PTI       : 3;
    Bit16 VCI_low   : 4;
};
#endif



// ATM UNI cell header as seen by an Intel processor
// after swapping the bytes using 80486/Pentium BSWAP
// instruction, or as seen by a big-endian CPU without
// any swapping
//
// remember that Intel bitfields are allocated
// from lsb to msb within a byte, but big-endian
// bitfields are allocated from high to low within
// a byte
//
// sorry, the Borland C++ 4.5 16-bit compiler really
// screws up when you want a bitfield to straddle
// more than one 16-bit boundary, so you need BCC32
//
struct AtmHeader
{
#if HOST_IS_BIG_ENDIAN
    Bit16 GFC   : 4;                    // generic flow control
    Bit16 VPI   : 8;                    // virtual path identifier
    Bit16 VCI   : 16;                   // virtual channel identifier
    Bit16 PTI2  : 1;
    Bit16 PTI1  : 1;
    Bit16 PTI0  : 1;
    Bit16 CLP   : 1;                    // cell loss priority
#else
    Bit16 CLP   : 1;                    // cell loss priority
    Bit16 PTI0  : 1;
    Bit16 PTI1  : 1;
    Bit16 PTI2  : 1;
    Bit16 VCI   : 16;                   // virtual channel identifier
    Bit16 VPI   : 8;                    // virtual path identifier
    Bit16 GFC   : 4;                    // generic flow control
#endif
};

// aliases for various bits of PTI
//
#define LAST        PTI0                // last cell in AAL5 PDU
#define CE          PTI1                // congestion experienced
#define OAM         PTI2                // this is an OAM F5 cell
#define E2E         PTI0                // OAM cell is end-to-end, not a segment
#define RESERVED    PTI1                // OAM cell is reserved for future use



// extract just VPI and VCI from (UNI) ATM header
//
// note that this accepts the 32-bit alias,
// rather than manipulating the bitfields,
// for speed
//
// #define VPI_VCI(x)   ((x.VPI << 16) | VCI(x))
//
#define VPI_VCI(x)   ((x >> 4) & 0xFFFFFF)



// extract just VPI and VCI from (NNI) ATM header
//
// note that this accepts the 32-bit alias,
// rather than manipulating the bitfields,
// for speed
//
// #define NNI_VPI_VCI(x)   ((x.GFC << 24) | (x.VPI << 16) | VCI(x))
//
// assumes that shifting right will shift zeroes
// into the most significant bits, not ones, and
// hence doesn't do the bitwise AND to clear those
//
// #define NNI_VPI_VCI(x)   ((x >> 4) & 0xFFFFFFF)
//
#define NNI_VPI_VCI(x)      (x >> 4)



union AtmHeaderCoerce
{
    Bit32 whole;
    struct AtmHeader bits;
};


#endif      // ifndef __ATM_H
