#include <stdio.h>                      // fopen()
#include <errno.h>                      // errno
#include <conio.h>                      // kbhit()
#include "..\types\types.h"             // Bit32 Bit8



// manifest constants
//
#define DISPLAY_MODULUS                 1000



// conditional compilation switches
//
#define ALLOW_KEYPRESS_TO_HALT          TRUE



// global variables
//
Bit8 bitswap[ 256 ];



int main( int argc, char *argv[] )
{
    const char *program = argv[ 0 ];
    const char *infile  = argv[ 1 ];
    const char *outfile = argv[ 2 ];
    FILE *in_handle;
    FILE *out_handle;
    int result;
    int low_nibble;
    int high_nibble;
    int garbage_char;
    Bit32 byte_count = 0;
    int one_byte;                       // mem corruption if not wide enough!
    int b0, b1, b2, b3, b4, b5, b6, b7;

    if( argc!=3 && argc!=4 )
    {
        fprintf( stderr, "converts a file with uppercase or lowercase hex digits\n" );
        fprintf( stderr, "into binary bytes, ignoring all whitespace\n\n" );
        fprintf( stderr, "usage: %s infile outfile {bitswap}\n", program );
        return 1;
    }

    in_handle = fopen( infile, "rb" );
    if( in_handle == NULL )
    {
        fprintf( stderr, "error %d opening input file\n", errno );
        return 2;
    }

    out_handle = fopen( outfile, "wb" );
    if( out_handle == NULL )
    {
        fprintf( stderr, "error %d opening output file\n", errno );
        return 3;
    }

    // initialize the bitswapping lookup table
    //
    for( b7=0; b7<2; b7++ )
        for( b6=0; b6<2; b6++ )
            for( b5=0; b5<2; b5++ )
                for( b4=0; b4<2; b4++ )
                    for( b3=0; b3<2; b3++ )
                        for( b2=0; b2<2; b2++ )
                            for( b1=0; b1<2; b1++ )
                                for( b0=0; b0<2; b0++ )
                                    bitswap[ b7<<7 | b6<<6 | b5<<5 | b4<<4 | b3<<3 | b2<<2 | b1<<1 | b0 ]
                                    =        b0<<7 | b1<<6 | b2<<5 | b3<<4 | b4<<3 | b5<<2 | b6<<1 | b7;

    while( !ferror( in_handle )
#if ALLOW_KEYPRESS_TO_HALT
        && !kbhit()
#endif
    )
    {
        // keep trying to read the high hex nibble from the input file
        //
        // we do the nibbles separately so we can ignore
        // whitespace and garbage between them
        //
        while( 1 != fscanf( in_handle, "%1X", &high_nibble ) )
        {
            if( feof( in_handle ) )
                goto end_of_file;

            garbage_char = fgetc( in_handle );
            fprintf( stderr, "ignoring garbage '%c' while reading high nibble of byte %d from input file\n", garbage_char, byte_count+1 );
        }

        // keep trying to read the low hex nibble from the input file
        //
        while( 1 != fscanf( in_handle, "%1X", &low_nibble ) )
        {
            if( feof( in_handle ) )
                goto end_of_file;

            garbage_char = fgetc( in_handle );
            fprintf( stderr, "ignoring garbage '%c' while reading low nibble of byte %d from input file\n", garbage_char, byte_count+1 );
        }

        // reconstruct the byte from its 2 nibbles
        //
        one_byte = (high_nibble * 16) + low_nibble;

        // count the bytes output
        //
        byte_count++;

        // show some status to the user (occasionally)
        //
        if( byte_count % DISPLAY_MODULUS == 0 )
            fprintf( stderr, "%lu ", byte_count );

        // if the user wants me to...
        //
        if( argc == 4 )

            // swap the bits of the byte lsb-for-msb
            //
            one_byte = bitswap[ one_byte ];

        // write the binary byte to the output file
        //
        result = fputc( one_byte, out_handle );
        if( result == -1 )
        {
            fprintf( stderr, "error %d writing output file\n", errno );
            break;
        }
    }
end_of_file:
    fprintf( stderr, "\n" );

#if ALLOW_KEYPRESS_TO_HALT
    if( kbhit() )
    {
        getch();
        fprintf( stderr, "halted by keypress!\n" );
    }
#endif

    // show statistics
    //
    fprintf( stdout, "%lu bytes output\n", byte_count );

    fclose( out_handle );
    fclose( in_handle );

    // return successfully
    //
    return 0;
}

