#include <stdio.h>              // printf()
#include <stdlib.h>             // exit()
#include "phys.h"



typedef Bit32 (*reader_t)( void *pointer );



// this set of 3 functions takes care of reading
// a range of addresses using only the width specified
//
void copy32( Bit32 *dest, Bit32 *src, Bit32 count )
{
    while( count >= 16 )
    {
        *dest = *src;
        *(dest +  1) = *(src +  1);
        *(dest +  2) = *(src +  2);
        *(dest +  3) = *(src +  3);
        *(dest +  4) = *(src +  4);
        *(dest +  5) = *(src +  5);
        *(dest +  6) = *(src +  6);
        *(dest +  7) = *(src +  7);
        *(dest +  8) = *(src +  8);
        *(dest +  9) = *(src +  9);
        *(dest + 10) = *(src + 10);
        *(dest + 11) = *(src + 11);
        *(dest + 12) = *(src + 12);
        *(dest + 13) = *(src + 13);
        *(dest + 14) = *(src + 14);
        *(dest + 15) = *(src + 15);

        dest += 16;
        src += 16;
        count -= 16;
    }

    while( count-- )
        *dest++ = *src++;
}



void copy16( Bit16 *dest, Bit16 *src, Bit32 count )
{
    while( count-- )
        *dest++ = *src++;
}



void copy8( Bit8 *dest, Bit8 *src, Bit32 count )
{
    while( count-- )
        *dest++ = *src++;
}



// this set of 3 functions takes care of touching
// only the address and width that is wanted by
// the user
//
// no masking is necessary because C promotion
// will zero out the high bits if necessary
//
Bit32 read32( void *pointer )
{
    return * (Bit32 *) pointer;
}



Bit32 read16( void *pointer )
{
    return * (Bit16 *) pointer;
}



Bit32 read8( void *pointer )
{
    return * (Bit8 *) pointer;
}



int main( int argc, char *argv[] )
{
    Bit32 handle, start, line;
    Bit32 length, length_remaining;
    Bit32 value;
    int width;
    const char *program_name = argv[0];
    Bit8 *pointer, *address;
    Bit8 *copy_pointer;
    reader_t reader;

    // parse command-line arguments
    //
    if( argc != 4 )
    {
        fprintf( stderr, "uses a PCI burst to copy a range of addresses, then displays them\n" );
        fprintf( stderr, "usage: %s phys-address bit-width length-in-bytes\n", program_name );
        return 1;
    }
    start  = strtoul( argv[1], NULL, 0 );
    length = strtoul( argv[3], NULL, 0 );
    width  = atoi( argv[2] );

    switch( width )
    {
        default:
            fprintf( stderr, "Sorry, you can only use widths of 8, 16, or 32 bits.\n" );
            return 2;

        case 8:
            reader = read8;
            break;

        case 16:
            reader = read16;
            break;

        case 32:
            reader = read32;
            break;
    }

    // map one or more pages of memory starting at the
    // physical address that the user specified
    //
    pointer = (Bit8 *) map_physical( start, length, &handle );
    if( pointer == NULL )
    {
        printf( "Sorry, cannot map %d bytes at physical address 0x%.8X to a flat pointer\n", length, address );
        return 3;
    }

    // allocate host memory to copy into
    //
    copy_pointer = (Bit8 *) malloc( length );
    if( copy_pointer == NULL )
    {
        printf( "Sorry, cannot allocate %d bytes of host memory to copy into\n", length );
        return 4;
    }

    printf( "copying %d-bit memory values starting at address 0x%.8X, length 0x%.8X\n", width, start, length );
    switch( width )
    {
        case 8:
            copy8( copy_pointer, pointer, length );
            break;

        case 16:
            copy16( (Bit16 *)copy_pointer, (Bit16 *)pointer, length/2 );
            break;

        case 32:
            copy32( (Bit32 *)copy_pointer, (Bit32 *)pointer, length/4 );
            break;
    }

    printf( "dumping copy from host memory\n" );
    for( address=copy_pointer,length_remaining=length; length_remaining>15; address+=16,length_remaining-=16 )
    {
        printf( "%.8X  ", (address-copy_pointer)+start );

        for( line=0; line<16; line+=width/8 )
        {
            value = reader( address + line );
            printf( " %.*X", width/4, value );
        }

        printf( "\n" );
    }

    return 0;
}
