#include <stdio.h>              // printf()
#include <stdlib.h>             // exit()
#include <io.h>                 // close() open() write()
#include <fcntl.h>              // O_CREATE O_TRUNC O_BINARY O_WRONLY
#include <sys/stat.h>           // S_IREAD S_Iwrite
#include "phys.h"



#define BLOCK_SIZE      512



int main( int argc, char *argv[] )
{
    Bit32 handle, start;
    Signed32 length, length_expected;
    const char *program_name = argv[0];
    const char *file_name = argv[1];
    Bit8 *pointer;
    int file_handle;
    int result;

    // parse command-line arguments
    //
    if( argc != 4 )
    {
        fprintf( stderr, "usage: %s filename phys-address length\n", program_name );
        return 1;
    }
    start  = strtoul( argv[2], NULL, 0 );
    length = strtoul( argv[3], NULL, 0 );

    // map one or more pages of memory starting at the
    // physical address that the user specified
    //
    pointer = (Bit8 *) map_physical( start, length, &handle );
    if( pointer == NULL )
    {
        printf( "Sorry, cannot map %d bytes at physical address 0x%.8X to a flat pointer\n", length, start );
        return 2;
    }

    printf( "saving memory values starting at address 0x%.8X, length 0x%.8X\n", start, length );

    // create file with r/w directory attributes if it doesn't exist
    // truncate file if it already exists
    // writes to files will be binary (no newline translation)
    // during execution, we need only write access
    //
    file_handle = open( file_name, O_CREAT|O_TRUNC|O_BINARY|O_WRONLY, S_IWRITE|S_IREAD );
    if( file_handle < 0 )
    {
        printf( "error %d opening file '%s' for writing\n", errno, file_name );
        return 3;
    }

    // write an integral number of 512-byte blocks to disk
    //
    length_expected = (length/BLOCK_SIZE)*BLOCK_SIZE;
    result = write( file_handle, pointer, length_expected );
    if( result < 0 )
    {
        printf( "error %d writing even %d bytes at logical 0x%.8X to file '%s'\n", errno, length_expected, pointer, file_name );
        return 4;
    }

    // write the rest of the data to disk
    //
    pointer += length_expected;
    length_expected = length - length_expected;
    if( length_expected > 0 )
    {
        result = write( file_handle, pointer, length_expected );
        if( result < 0 )
        {
            printf( "error %d writing remaining %d bytes at virtual 0x%.8X to file '%s'\n", errno, length_expected, pointer, file_name );
            return 5;
        }
    }

    result = close( file_handle );
    if( result < 0 )
    {
        printf( "error %d closing file '%s'\n", errno, file_name );
        return 6;
    }

    return 0;
}
