/* conditional compile switches all go here
 *
 * manifest constants are still sprinkled
 * all through the code (sorry)
 */
#define STUB_0TH_LEVEL_STATS    FALSE           // just for debugging
#define STUB_2ND_LEVEL_STATS    FALSE           // just for debugging
#define STUB_DO_RECORD          FALSE           // just for debugging
#define STUB_FLOWS_EXPIRE       FALSE           // just for debugging
#define FORCE_1_BUCKET          FALSE           // just for debugging
#define FORCE_1_FLOW_PER_BUCKET FALSE           // just for debugging
#define ASSUME_OBJECT_IS_PQ     TRUE
#define ASSUME_ETHERTYPE_IS_IP  FALSE
#define TIME_ALREADY_SWAPPED    TRUE
#define BYTESWAP_FIELDS_EARLY   FALSE
#define ALWAYS_DUMP_HASH_INFO   FALSE
#define USE_OLD_HASH            FALSE
#define FOLD_IS_JUST_MODULUS    FALSE
#define FOLD_RANDOMIZES         TRUE
#define FOLD_KEEPS_16_BITS      TRUE
#define DEBUG_CHAIN_LENGTH      FALSE
#define BUFFER_USED_PARANOIA    FALSE
#define BUFFER_PTR_PARANOIA     FALSE
#define FLOAT_EXCEPT_PARANOIA   FALSE
#define HASHINDEX_PARANOIA      FALSE
#define SUMMARIZE_ONE_PARANOIA  FALSE
#define DEBUG_HASHINDEX         FALSE
#define DEBUG_COPY_CELL         FALSE           // stupid debugger!
#define DEBUG_FILLIN            FALSE
#define DEBUG_HOUSECLEAN        FALSE
#define FAIL_ON_PACKET_COUNT    FALSE
#define TALK_WHEN_ALLOC_FAILS   FALSE           // for debugging when overflow occurs only!  too slow
#define DELTA_TIME_PARANOIA     FALSE           // worry about negative delta time since last flows_expire()
#define DELTA_TIME_FIXUP        FALSE           // truncate large time since last flows_expire() to 1 second
#define DELTA_TIME_STATS        TRUE            // track min/max/mean for delta time since last flows_expire()
#define DEBUG_QUERY_TOTALS      TRUE
#define DEBUG_MAX_BPS           FALSE
#define MAX_BPS_PARANOIA        TRUE
#define QUERY_CALLS_EXPIRE      FALSE
#define DEBUG_ACTIVE_FEED       FALSE
#define DEBUG_ACTIVE_FEED_DATA  FALSE
#define EXPIRE_OVERHEAD_STATS   TRUE
#define REPORT_OVERHEAD_STATS   FALSE
#define DEBUG_EXTRA_TCP_STATS   FALSE
#define DEBUG_REASSEMBLY        FALSE
#define DEBUG_ORPHAN_FRAGS      FALSE
#define DEBUG_OVERLAP_FRAGS     FALSE
#define DEBUG_HOLE_FRAGS        FALSE
#define DEBUG_NON_IP            TRUE
#define DEBUG_NON_IP_PER_VPI    TRUE
#define COUNT_NON_IP_PER_VPI    TRUE
#define ASSUME_NNI_ADDRESSING   FALSE
#define RESELECTABLE_FLOWTRACE  TRUE
#define DEBUG_BLOCK_LOCKUP      FALSE
#define DEBUG_BLOCK_LOCKUP_2    FALSE

#ifdef DOS  /* For NeTraMet */
#define SHOW_CPU_UTILIZATION    FALSE  
#else
#define SHOW_CPU_UTILIZATION    TRUE
#endif



#include <sys/types.h>

#if __unix__
#   include <sys/time.h>
#   include <sys/resource.h>
#endif
#include "../wattcp/tcp.h"

#include <stdio.h>
#include <stdlib.h>             /* atoi() getenv() min() ?others? */
#include <fcntl.h>
#include <errno.h>
#include <float.h>              /* DBL_MIN DBL_MAX */

#include "../types/types.h"
#if !__unix__
#   include "..\intel\timestmp.h"
#   include "..\intel\byteswap.h"
#   include "..\intel\fpu.h"    /* get_fpu_status() clear_fpu_status() */
#endif

#include "..\integrat\tcp_seq.h"            /* SEQ_LT() */
#include "..\integrat\queue.h"
#include "..\integrat\trace.h"              /* must be included before util.h */
#include "..\integrat\util.h"
#include "..\integrat\dostats.h"
#include "..\integrat\buffer.h"             /* class BufferBlock */
#include "..\integrat\fs2flows.h"           /* make sure we comply with our own prototypes */
#include "..\integrat\portstat.h"           /* porthash_alloc_failures      (that module should print this itself, but I don't want to break the Perl scripts just now) */
#include "..\integrat\hoststat.h"           /* hosthash_alloc_failures      (that module should print this itself, but I don't want to break the Perl scripts just now) */
#include "..\integrat\netstats.h"           /* nethash_alloc_failures       (that module should print this itself, but I don't want to break the Perl scripts just now) */
#include "..\integrat\fragstat.h"           /* fraghash_alloc_failures      (that module should print this itself, but I don't want to break the Perl scripts just now) */



// none of this code gets compiled in
// if it's isn't going to be called
//
#if DO_FOREGROUND_STATS



#if SHOW_CPU_UTILIZATION
extern Bit32 blocks;
extern PentiumClock before, elapsed;
extern PentiumClock startup;
extern PentiumClock min_elapsed, max_elapsed, sum_elapsed;
#endif




// called in foreground by main data capture program
//
int flows_per_block( BufferBlock *block )
{
    int final_result = 0;
    int result;
    unsigned count, max_count;
    Bit8 interface = block->cardindex;
    tracerec *record;
    union { FpuStatus float_status; Bit16 float_status_alias; };
#if DEBUG_COPY_CELL
    tracerec debug_record;
#endif

#if BUFFER_USED_PARANOIA
    // safety check: used count cannot be greater
    // than COMPL_RING_ENTRIES
    //
    if( block->used > COMPL_RING_ENTRIES )
    {
        fprintf( stderr, "DANGER!   block used %d > %d max!\n", block->used, COMPL_RING_ENTRIES );
        block->used = COMPL_RING_ENTRIES;
        blocks_overrun++;
    }
#endif

#if SHOW_CPU_UTILIZATION
    // start the timer of flows processing
    //
    before.set();
#endif

    // for all packets headers in the block
    //
    // we cannot process the last cell in the block if we didn't get this
    // block by taking it off the filled list, because cells_reassemble()
    // cannot safely free the block, so we want to guarantee that we will
    // call cells_reassemble() at least once later, when we see this block
    // legitimately
    //
    // note that since we do not check the block's used field inside the loop,
    // we will not notice cells that come in during the loop, but hey, the only
    // way the interrupt handler could be called to update that count would be
    // if the block got full, and in that case we have to revisit the block
    // anyways (see comment immediately above)
    //
    max_count = block->used;
    if( block->state != is_consuming
    &&  max_count > 0 )
        max_count--;
    for( count=block->queried; count<max_count; count++ )
    {
#if DEBUG_BLOCK_LOCKUP
        fprintf( stderr, "%5d\b\b\b\b\b", count );
#endif
#if DEBUG_BLOCK_LOCKUP_2
        if( count == reccount_trace )
            fprintf( stderr, "debugger wants to stop here on block's record %d\n", count );
#endif

#if BUFFER_PTR_PARANOIA
        // safety check: buffer cannot be before
        // the start of the flat memory pool
        //
        if( block->buffers < flat_pool )
        {
            printf( "danger!  corrupted bufferblock!  stop here in debugger\n" );
            return ERROR_FLOWS_CORRUPTED_BUFFERBLOCK;
        }
#endif

#if FLOAT_EXCEPT_PARANOIA
        // if there was any exception but precision...
        //
        float_status_alias = get_fpu_status();
        if( float_status_alias & 0xDF )
            printf( "before record %d floating point status is 0x%.4X\n", count, float_status_alias );
        clear_fpu_status();
#endif

        // we must alter the block's queried count in the loop before calling
        // cells_reassemble(), because if that triggers freeing this block,
        // and a card's interrupt handler allocates the same block before we
        // try to alter the block's queried count after the loop (made more
        // probable by the LIFO nature of the free list, but still quite
        // unlikely since there are so few instructions between those 2 steps),
        // we would actually be altering the queried count of some other block
        // which we don't own, causing some or all of its packets to never
        // be processed!
        //
        // this is also why we cannot use block->queried directly as the
        // loop variable in the for() statement, since it increments after
        // the last iteration
        //
        block->queried++;

        record = (tracerec *) (block->buffers + count*BUFFER_SIZE);

#if DEBUG_COPY_CELL
        // copy the record because the debugger, for some
        // idiotic reason does not have the same addressibility
        // that our program has
        //
        debug_record = *record;
#endif

#if STUB_0TH_LEVEL_STATS
        cellhashrec chr;
        chr.first = record;
        chr.second = NULL;
        chr.last = (tracerec_last *) record;
        chr.interface = interface;
#if !CARD_SWAPS_ATM_HEADER
        record->atm.whole = ntohl( record->atm.whole );
#endif
        result = flows_per_packet( &chr, interface );
#else
        // this might trigger freeing this or some other block,
        // depending upon how many cells per packet the card is
        // capturing, whether a packet straddles blocks, and whether
        // this is the last cell in this block
        //
        result = cells_reassemble( record, interface );
#endif
        if( final_result == 0 )
            final_result = result;

    }   // end for all records in block

#if DEBUG_REASSEMBLY
    cells_dump_all_packets();
#endif

#if STUB_0TH_LEVEL_STATS
    // if the card isn't still filling the block...
    //
    if( block->state == is_consuming )
    {
        // release this block, since we are sure
        // all cells within it are unused now
        //
        result = all_buffers->free( block );
        if( result && !final_result )
            final_result = result;
    }
    else
        if( block->state != is_filling )
            fprintf( stderr, "flows_per_block() warning: block state is %d, not consuming=%d or filling=%d\n", block->state, is_consuming, is_filling );
#endif

#if SHOW_CPU_UTILIZATION
    // stop the timer of flow processing
    //
    elapsed.set();

    // update the flow processing time statistics
    //
    blocks++;
    elapsed -= before;
    sum_elapsed += elapsed;
    if( max_elapsed < elapsed )
        max_elapsed = elapsed;
    if( min_elapsed > elapsed )
        min_elapsed = elapsed;
#endif

    return final_result;
}

#endif /* DO_FOREGROUND_STATS */