#include <stdio.h>              /* printf() */
#include <conio.h>              /* kbhit() getch() */
#include <stdlib.h>             /* atoi() */
#include "../types/types.h"     /* TRUE FALSE */



// conditional compile switches
//
#define OPTIMIZE_FOR_INTEL  FALSE           // doesn't work!



// manifest constants
//
#define MAX_PRIMES          10000000
#define FACTOR_LOOP_LIMIT   1000
#define PRIMES_PER_LINE     10



// global variables
//
unsigned primes[ MAX_PRIMES ];
unsigned prime_count;



/* simple sieve of erastosthenes prime number generator
 *
 * can also find the prime factors of a number user supplies
 */

int main( int argc, char *argv[] )
{
    unsigned count;
    unsigned testee;
    unsigned composite;
    unsigned quotient;
    int keystroke;
    int result;
    int is_prime;
    const char *program = argv[ 0 ];
    const char *filename = argv[ 1 ];
    const char *composite0 = argv[ 2 ];
    FILE *handle = NULL;

    if( argc > 3 )
    {
        fprintf( stderr, "usage: %s {file-with-primes} {number-to-factor}\n", program );
        return 1;
    }

    if( argc >= 2 )
    {
        // try to open the file named
        //
        handle = fopen( filename, "r" );
        if( handle == NULL )
        {
            fprintf( stderr, "cannot open file '%s' for reading\n", filename );
            return 2;
        }
    }

    if( argc == 3 )
    {
        // try to convert the number from a string
        //
        composite = atoi( composite0 );
        if( composite <= 1 )
        {
            fprintf( stderr, "number to factor must be greater than 1\n" );
            return 3;
        }
        fprintf( stderr, "trying to factor %u\n", composite );
    }

    for( prime_count=0,testee=1; prime_count<MAX_PRIMES; prime_count++ )
    {
        if( handle )
        {
            // read a prime from the input file
            //
            result = fscanf( handle, "%d", &testee );
            if( result <= 0 )
            {
                // close the file
                //
                fclose( handle );
                handle = NULL;
                fprintf( stderr, "read %d primes from file '%s'\n", prime_count, filename );
            }
        }

        if( !handle ) do
        {
            testee++;
            is_prime = TRUE;

            for( count=0; count<prime_count; count++ )
            {
                // note: Intel processors calculate both the remainder
                // and quotient with one instruction, but the stupid
                // compiler will probably generate 2 of them
                //
                if( testee % primes[count] == 0 )
                {
                    is_prime = FALSE;
                    break;
                }
#if OPTIMIZE_FOR_INTEL
                // quotient always ends up in _EAX register
                //
                quotient = _EAX;
                if( quotient < primes[count] )
#else
                if( testee / primes[count] < primes[count] )
#endif
                    break;
            }

        } while( !is_prime );

        if( prime_count % PRIMES_PER_LINE )
            printf( "%d ", testee );
        else
            printf( "%d\n", testee );

        primes[ prime_count ] = testee;

        if( argc == 3 )
        {
            for( count=0; count<FACTOR_LOOP_LIMIT; count++ )
            {
                if( composite % testee != 0 )
                    break;
                else
                    composite /= testee;
            }
            if( count > 0 )
            {
                fprintf( stderr, "%d is a factor %d times, leaving %d\n", testee, count, composite );
            }

            // this and all smaller primes are no longer factors of remaining
            // composite, so any other prime factors must leave a quotient
            // that is larger than this prime, which limits how large a prime
            // could possibly be a factor to (composite/thisprime)
            //
            if( testee * testee > composite )
            {
                fprintf( stderr, "tested %d primes up %d as factors\n", prime_count, testee );
                break;
            }
        }

        if( kbhit() )
        {
            fprintf( stderr, "primes[%d] = %d\n", prime_count, testee );

            keystroke = getch();
            if( keystroke == 27 )
                break;
        }
    }

    return 0;
}
