#ifndef _I8254_H
#define _I8254_H



// hardware timer chip, Intel part# 8254-2
//
// actually, this code only controls the third
// timer of 3 on the chip, which usually controls
// the PC speaker
//
// only one instance of this class is allowed per
// computer, since there is only one of the hardware
// it models
//
// how do I guarantee that???
//
class HardwareTimer
{

public:
    // constructor
    //
    HardwareTimer()
    {
        // tell the 8254-2 chip that we want timer 0
        // (bits 7,6 = 00)
        //
        // to count in binary (bit 0 = 0)
        //
        // and that we want to access its LSB followed by
        // MSB on two successive reads or writes to its port
        // (bits 5,4 = 11)
        //
        // note: if we wanted to just access its MSB with a
        // single access to its port, bits 5,4 should be 10
        //
        // and that we want it to act as a rate generator
        // (output high except for one clock when it
        // decrements to zero, count reloads automatically)
        // (bits 3,2,1 = 010 or 110)
        //
        outportb( 0x43, 0x34 );
    }

    // destructor
    //
    ~HardwareTimer()
    {
        // tell the 8254-2 chip that we want timer 0
        // (bits 7,6 = 00)
        //
        // to count in binary (bit 0 = 0)
        //
        // and that we want to access its LSB followed by
        // MSB on two successive reads or writes to its port
        // (bits 5,4 = 11)
        //
        // and that we want it to
        // (decrement in two's, interrupt every second cycle,
        // count reloads automatically)
        // (bits 3,2,1 = 110)
        //
        outportb( 0x43, 0x36 );
    }

    // set the count of the device, in units of
    // 840.3361345 nanoseconds (1 / 1.19 MHz)
    //
    // you might want to disable interrupts around this
    // but it would be unsafge for me to do so because
    // I can't tell if they are enabled already!
    //
    void set( unsigned value )
    {
        unsigned short old_flags = _FLAGS;

        // disable maskable interrupts
        //
        disable();

        // set the least significant byte of timer 0
        //
        outportb( 0x40, value & 0xFF );

        // set the most significant byte of timer 0
        //
        outportb( 0x40, value >> 8 );

        // if interrupts used to be enabled...
        //
        if( old_flags & 0x200 )

            // enable maskable interrupts again
            //
            enable();
    }

    // get the current counter value, in units of
    // 840.3361345 nanoseconds (1 / 1.19 MHz)
    //
    unsigned get( void )
    {
        unsigned value;
        unsigned short old_flags = _FLAGS;

        // disable interrupts
        //
        disable();

        // tell timer 0 to latch its current count so
        // it doesn't change while we are reading it
        //
        outportb( 0x43, 0x00 );

        // get the least significant byte of timer 0
        //
        value = inportb( 0x40 );

        // get the most significant byte of timer 0
        //
        value |= inportb( 0x40 ) << 8;

        // if interrupts used to be enabled...
        //
        if( old_flags & 0x200 )

            // enable maskable interrupts again
            //
            enable();

        return value;
    }
};


#endif      // not included yet
