#include <stdio.h>              // printf()
#include <stdlib.h>             // exit()
#include "pci.h"



int main( int argc, char *argv[] )
{
    Bit16 handle;
    Bit16 offset;
    Bit32 old_value, new_value;
    int width;
    int want_write;
    const char *program_name = argv[0];

    // parse command-line arguments
    //
    if( argc<4 || argc>5 )
    {
        fprintf( stderr, "usage: %s card-handle config-offset bit-width {config-value}\n", program_name );
        return 1;
    }
    handle = strtoul( argv[1], NULL, 0 );
    offset = strtoul( argv[2], NULL, 0 );
    width = atoi( argv[3] );
    if( argc == 5 )
    {
        new_value = strtoul( argv[4], NULL, 0 );
        want_write = TRUE;
    }
    else
        want_write = FALSE;

    // check whether PCI BIOS is present
    //
    if( !pci_is_bios_present() )
    {
        printf( "Sorry, PCI BIOS is not present.\n" );
        return 2;
    }

    switch( width )
    {
        default:
            fprintf( stderr, "Sorry, you can only use widths of 8, 16, or 32 bits.\n" );
            return 3;

        case 8:
            old_value = pci_read_config_8( handle, offset );
            if( want_write )
                pci_write_config_8( handle, offset, new_value );
            break;

        case 16:
            old_value = pci_read_config_16( handle, offset );
            if( want_write )
                pci_write_config_16( handle, offset, new_value );
            break;

        case 32:
            old_value = pci_read_config_32( handle, offset );
            if( want_write )
                pci_write_config_32( handle, offset, new_value );
            break;
    }

    printf( "dealing with PCI handle 0x%.4X configuration space %d bits at a time\n", handle, width );
    printf( "read old value 0x%.*X at offset 0x%.4X\n", width/4, old_value, offset );

    if( want_write )
        printf( "wrote new value 0x%.*X\n", width/4, new_value );

    return 0;
}
