;        .RADIX  16
        TITLE   Stuffs scancodes into the keyboard controller HW
        SUBTTL  usable on IBM AT computers or better
        SUBTTL  should work with games that take over INT 9
        NAME    KBSTUFF



; compile-time switches
;
also_release_key        equ     0
checkpoint_in_year      equ     0



;----------------------------------------
;                                       |
; 8042 IBM AT keyboard controller ports |
;                                       |
;----------------------------------------



; output the first byte of commands here
; when KBOUTPUTBIT is cleared
;
KBCMDPORT       equ     64h

; read here any time to determine controller status
;
KBSTSPORT       equ     64h

; bits within KBSTSPORT
;
KBPARITYBIT     equ     10000000b       ;keyboard sent bad (even) parity
KBRECEIVEBIT    equ     01000000b       ;timeout while receiving from keyboard
KBTRANSMITBIT   equ     00100000b       ;timeout while transmitting to keyboard
KBINHIBITBIT    equ     00010000b       ;keyboard is inhibited
KBCOMMANDBIT    equ     00001000b       ;computer last wrote to KBCMDPORT, not KBDATAPORT
KBSYSTEMBIT     equ     00000100b       ;computer reset was caused by keyboard controller
KBINPUTBIT      equ     00000010b       ;input buffer full
KBOUTPUTBIT     equ     00000001b       ;output buffer full

; write data to be sent to the keyboard (not the controller)
; also write the second byte of a multibyte keyboard controller
;    command (such as KBWRITECMD)
; also read translated scancodes from the keyboard here
; also read the result of a keyboard controller command here
;
; can only write here when KBINPUTBIT  bit in KBSTSPORT is 0
; can only read  here when KBOUTPUTBIT bit in KBSTSPORT is 1
;
KBDATAPORT      equ     60h



; read the keyboard controller's command byte at KBDATAPORT
;
KBREADCMD       equ     20h

; write the keyboard controller's command byte from the
; next byte written to KBDATAPORT
;
KBWRITECMD      equ     60h

; perform internal diagnostic tests and place 0x55
; in output buffer if successful
;
KBSELFTESTCMD   equ     0AAh

; test keyboard clock and keyboard data lines
; result codes:
;     00    no error detected
;     01    keyboard clock stuck low
;     02    keyboard clock stuck high
;     03    keyboard data stuck low
;     04    keyboard data stuck high
;
KBIFTESTCMD     equ     0ABh

; dump 16 bytes of keyboard controller RAM,
; ?then current input port state,
; then current output port state
; then controller's program status word
;
; all are sent in scancode format (what does that mean?)
; ?to output port?
;
KBDIAGDUMPCMD   equ     0ACh

; sets KBINHIBIT bit of command byte
;
KBDISABLECMD    equ     0ADh

; clears KBINHIBIT bit of command byte
;
KBENABLECMD     equ     0AEh

; read keyboard controller's input port
; and place result in output buffer
; (should only use if output buffer is empty)
;
KBREADINPCMD    equ     0C0h

; read keyboard controller's output port
; and place result in output buffer
; (should only use if output buffer is empty)
;
KBREADOUTPCMD   equ     0D0h

; next byte written to KBDATAPORT will be
; written to keyboard controller's output port
;
KBWRITEOUTPCMD  equ     0D1h

; read keyboard controller's test inputs
; to the output buffer
;
; bit 0 will have T0 input (A.K.A. keyboard clock input)
; bit 1           T1                        data input
;
KBREADTESTCMD   equ     0E0h

; pulse any of the low 4 bits of the
; keyboard controller's output port
;
; for each of the low 4 bits of this command value,
; if the bit is cleared, the corresponding bit of
; output port will be pulsed low for approximately
; 6 microseconds
;
; so (KBPULSECMD OR 00001110b) = 0FEh will pulse
; only the least significant bit of the output port
; which will reset the CPU
;
KBPULSEOUTPCMD  equ     0F0H
KBRESETCMD      equ     0FEh            ;actually a specific instance of KBPULSEOUTPCMD



; bits within the keyboard controller's command byte
;
KBRESERVED7     equ     10000000b       ;send as zero
KBTRANSLATE     equ     01000000b       ;translate keyboard scancodes to IBM PC
KBNOTRANSLATE   equ     00100000b       ;XT KB attached: don't check parity or translate
KBINHIBIT       equ     00010000b       ;force keyboard clock comm line negative
KBNOINHIBIT     equ     00001000b       ;disable inhibit mechanism
KBSYSTEMFLAG    equ     00000100b       ;remember that BIOS is about to reset CPU
KBRESERVED1     equ     00000010b       ;send as zero
KBINTERRUPT     equ     00000001b       ;interrupt computer when KBOUTPUTBIT is set



; bits within the keyboard controller's output port
;
KBDATAOUT       equ     10000000b       ;data output to keyboard
KBCLOCKOUT      equ     01000000b       ;clock output to keyboard
KBINBUFEMPTY    equ     00100000b       ;keyboard's input buffer is empty
KBOUTBUFFULL    equ     00010000b       ;keyboard's output buffer is full
KBRESERVED3     equ     00001000b       ;send as zero
KBRESERVED2     equ     00000100b       ;send as zero
KBGATEA20       equ     00000010b       ;A20 is enabled
KBSYSRESET      equ     00000001b       ;please reset CPU




; when a key is released, the AT keyboard itself (not to be confused
; with its controller on the motherboard) actually sends the 2-byte
; sequence "0F0h, original scancode"
;
; but the XT keyboard sends the one-byte sequence "original scancode
; OR'ed with 80h", and the keyboard controller on AT motherboards
; can be programmed (by setting the KBTRANSLATE bit of its command
; byte, see above) to emulate that behavior
;
; so you need to know what mode the motherboard keyboard controller
; is in, in order to know what break code the BIOS is expecting
;
SCANBREAK       equ     0F0h





io_delay   macro
    if 1
        push cx
        mov cx,65535d
        loop $
        pop cx
    else
    if 0
        jmp $+2
        jmp $+2
        jmp $+2
        jmp $+2
        jmp $+2
        jmp $+2
        jmp $+2
        jmp $+2
        jmp $+2
        jmp $+2
        jmp $+2
        jmp $+2
        jmp $+2
        jmp $+2
        jmp $+2
        jmp $+2
        jmp $+2
        jmp $+2
        jmp $+2
        jmp $+2
        jmp $+2
        jmp $+2
        jmp $+2
        jmp $+2
        jmp $+2
        jmp $+2
        jmp $+2
        jmp $+2
        jmp $+2
        jmp $+2
        jmp $+2
        jmp $+2
        jmp $+2
        jmp $+2
        jmp $+2
    endif
    endif
        endm



    if 0
demo_sysflag_off proc near

        cli

        ; read the keyboard controller's current command
        ; byte and save it in BL
        ;
        mov ah, KBREADCMD
        call writecmd
        call readdata
        mov bl, al

        ; tell keyboard controller to turn off the system flag
        ;
        mov ah, KBWRITECMD
        call writecmd
        mov ah, bl
        and ah, NOT KBSYSTEMFLAG
        call writedata

        ; wait for system flag to be off at status port
        ;
test1_1:
        in al, KBSTSPORT
        test al, 4
        jnz test1_1

        sti

        ret

demo_sysflag_off endp
    endif



    if 0
demo_diag_dump proc near

        ; tell keyboard controller to do diagnostic dump
        ;
        mov ah, KBDIAGDUMPCMD
        call writecmd

        ; read lots of scancodes into memory
        ;
        mov si, offset scanbuf1
        mov cx, offset scanend1
        sub cx, si
        call readmany

        ; exit program
        ;
        ret

demo_diag_dump endp
    endif



    if 0
demo_stuffmany proc near

        ; point at the first scan code
        ;
        mov si, offset scanbuf2

        ; calculate how many scancodes there are
        ;
        mov cx, offset scanend2
        sub cx, si

        ; stuff 'em all!
        ;
        call stuffmany

        ; exit program
        ;
        ret

demo_stuffmany endp
    endif



    if 0
demo_kb_echo proc near

        ; disable interruts
        ;
        cli

        ; send command to keyboard (not the controller
        ; for it on the motherboard) to echo
        ;
        ; unfortunately, it echoes the command itself
        ; and does not expect another byte to follow,
        ; so we cannot use this to echo scancodes back
        ; to us
        ;
        mov ah, 0EEh
        call writedata

        ; read 1st response from keyboard
        ;
        call readdata
        mov bl, al

        ; read 2nd response from keyboard
        ;
        call readdata
        mov bh, al

        ; enable interrupts again
        ;
        sti

        ; return to caller
        ;
        ret

demo_kb_echo endp
    endif




; set LED indicators on keyboard
;
; input bits in BH are as follows:
;       0   scroll lock
;       1   num lock
;       2   caps lock
;
; output: BL = first KB response
;         BH = seconds KB response
;
setlights proc near

        ; disable interruts
        ;
        cli

        ; send command to keyboard (not the controller
        ; for it on the motherboard) to set LED
        ;
        mov ah, 0EDh
        call writedata

        ; read 1st response from keyboard
        ;
        ; should be ACK (0FAh) not RESEND (0FEh)
        ;
        call readdata
        mov bl, al

        ; send the LED state desired to keyboard
        ;
        mov ah, bh
        call writedata

        ; read 2nd response from keyboard
        ;
        ; should be ACK (0FAh) not RESEND (0FEh)
        ;
        call readdata
        mov bh, al

        ; enable interrupts again
        ;
        sti

        ; return to caller
        ;
        ret

setlights endp




; input:    address in DH
;
; output:   value in AL
;
; affects:  DH
;
readrtc proc near

        pushf
        cli

        ; preserve NMI mask and output the address
        ;
        in al, 70h
        and al, 80h
        and dh, 7Fh
        or al, dh
        out 70h, al

        ; input the data
        ;
        mov al, dl
        in al, 71h

        popf
        ret

readrtc endp





; input:    address in DH, value in DL
;
; affects:  AL, DH
;
writertc proc near

        pushf
        cli

        ; preserve NMI mask and output the address
        ;
        in al, 70h
        and al, 80h
        and dh, 7Fh
        or al, dh
        out 70h, al

        ; output the data
        ;
        mov al, dl
        out 71h, al

        popf
        ret

writertc endp





; input:    pointer to scancodes in DS:SI
;           count of scancodes in CX
;
; alters:   AX, BX, CX, SI
;
; output:   none
;
stuffmany proc near

    if checkpoint_in_year
        ; use the realtime clock's year byte to hold our checkpoint
        ; in case the machine locks up during this subroutine
        ;
        push dx
        mov dh, 09h

        ; read the old value so we can see it in the debugger
        ;
        call readrtc

        ; start the checkpoint for this run at zero
        ;
        xor dl, dl
    endif

stuffmany_more:

    if checkpoint_in_year
        ; increment the high nibble of checkpoint
        ;
        add dl, 10h

        ; set low nibble of checkpoint to 0
        ;
        and dl, 0f0h

        ; save it in RTC
        ;
        call writertc
    endif

    if checkpoint_in_year
        ; set low nibble of checkpoint to 1
        ;
        and dl, 0f0h
        or dl, 1
        call writertc
    endif

        ; stuff the current scancode
        ;
        mov bh, [si]
        call stuffone

    if also_release_key
        ; let go of that key
        ;

    if checkpoint_in_year
        ; set low nibble of checkpoint to 2
        ;
        and dl, 0f0h
        or dl, 2
        call writertc
    endif

        ; if KB controller emulating XT...
        ;
        test bl, KBTRANSLATE
        jz stuffmany_not_emulating

        ; send the same scancode with its msb set
        ;
        or bh, 80h
        call stuffone
        jmp stuffmany_next

stuffmany_not_emulating:
        ; send the break code then the same scancode,
        ; as a 2-byte sequence
        ;
        mov bh, SCANBREAK
        call stuffone
        mov bh, [si]
        call stuffone
stuffmany_next:
    endif

        ; move on to the next scancode
        ;
        inc si

        ; check if we ran out of data
        ;
        dec cx
        jnz stuffmany_more

        ; return to the caller
        ;
        ret

stuffmany endp





; wait for motherboard keyboard controller's
; output buffer to be emptied by INT9 (IRQ1)
; interrupt handler
;
; assumes:  interrupts enabled
;
; input:    none
;
; alters:   AL
;
stuffwait proc near

        ; read the keyboard controller status port
        ;
        in al, KBSTSPORT
        io_delay

        ; wait for the output buffer to be empty
        ;
        and al, KBOUTPUTBIT
        jnz stuffwait

        ; return to caller
        ;
        ret

stuffwait endp





; input:    scancode to stuff in BH
;
; assumes:  interrupts are disabled
;
; alters:   AX, BX
;
; output:   KB controller's command byte in BL
;
; notes:    try as I might, I simply could not keep this code
;           from locking up my SuperMicro P5STE Pentium motherboard
;           eventually, even though it worked perfectly forever on
;           my NEC Versa 486/25 laptop, so it must be regarded
;           as quite unsafe!
;
stuffone proc near

        ; remember whether the caller had interrupts enabled
        ;
        pushf

        ; interrupts must be disabled while we play
        ; fast and loose with motherboard KB controller
        ;
        cli

        ; read the keyboard controller's current command
        ; byte and save it in BL register
        ;
        mov ah, KBREADCMD
        call writecmd
        call readdata
        mov bl, al

        ; set the command byte to be the scancode we want to stuff
        ;
        mov ah, KBWRITECMD
        call writecmd
        mov ah, bh
        call writedata

        ; tell the keyboard controller to put the command
        ; byte in the output buffer, but do not read it
        ;
        ; the INT 9 handler will process it when we
        ; re-enable interrupts
        ;
        mov ah, KBREADCMD
        call writecmd

        ; restore the previous command byte
        ;
        mov ah, KBWRITECMD
        call writecmd
        mov ah, bl
        call writedata

        ; tell the keyboard controller to inhibit the keyboard
        ;
        mov ah, KBDISABLECMD
        call writecmd

        ; maskable interrupts must be enabled for normal
        ; BIOS KB handler to execute
        ;
        sti

        call stuffwait

        ; if the keyboard didn't used to be inhibited...
        ; (not implemented yet -- always uninhibits)
        ;

        ; disable interrupts again, while we talk to the keyboard
        ; controller one last time
        ;
        cli

        ; tell the keyboard controller to stop inhibiting the keyboard
        ;
        mov ah, KBENABLECMD
        call writecmd

        ; leave interrupts the way we found them
        ;
        popf

        ; return to the caller
        ;
        ret

stuffone endp





writecmd proc near

        ; read the keyboard controller status port
        ;
        in al, KBSTSPORT
        io_delay

        ; wait for the input buffer to be empty
        ;
        and al, KBINPUTBIT
        jnz writecmd

        ; send the caller's command out the command port
        ;
        mov al, ah
        out KBCMDPORT, al
        io_delay

        ; return to the caller
        ;
        ret

writecmd endp




writedata proc near

        ; read the keyboard controller status port
        ;
        in al, KBSTSPORT
        io_delay

        ; wait for the input buffer to be empty
        ;
        and al, KBINPUTBIT
        jnz writedata

        ; send the caller's byte out the data port
        ;
        mov al, ah
        out KBDATAPORT, al
        io_delay

        ; return to the caller
        ;
        ret

writedata endp




readdata proc near

        ; read the keyboard controller status port
        ;
        in al, KBSTSPORT
        io_delay

        ; wait for the output buffer to be full
        ;
        and al, KBOUTPUTBIT
        jz readdata

        ; read the data port
        ;
        in al, KBDATAPORT
        io_delay

        ; return to the caller
        ;
        ret

readdata endp




; input:    pointer to buffer for scancodes in DS:SI
;           count of bytes available in CX
;
; alters:   AL
;
; output:   none
;
readmany proc near

        ; don't allow normal BIOS keyboard interrupt
        ; handler to execute
        ;
        cli

        ; wait for one scancode
        ;
        call readdata

        ; save the scancode in memory
        ;
        mov [si], al

        ; point at next available location
        ;
        inc si
        dec cx

        ; do it again if there is any room left
        ;
        jnz readmany

        ; allow normal interrupts again
        ;
        sti

        ; return to the caller
        ;
        ret

readmany endp



; sequence of scancodes read from keyboard port
;
scanbuf1    db 20h dup( ? )
scanend1    equ $

;        .RADIX  10
