#include <stdio.h>                      // printf()
#include <mem.h>                        // memset()
#include <stdlib.h>                     // strtoul()
#include "../scsi/aspi.h"               // AspiSrb aspi
#include "../intel/byteswap.h"          // htonl()
#include "../dpmi32/realmode.h"         // real_to_physical()



#define BLOCK_SIZE      512
#define TIMEOUT         100.0



Bit8 my_data[ BLOCK_SIZE ];



int main( int argc, char *argv[] )
{
    AspiSrb my_srb;
    AspiExecuteScsiIo *my_scsi_io = &my_srb.ExecuteScsiIo;
    ScsiWriteSame *my_write_same = &my_scsi_io->CdbAndSense.Cdb.WriteSame;
    unsigned count;
    const char *program = argv[ 0 ];
    const char *offset_string = argv[ 1 ];
    const char *blocks_string = argv[ 2 ];
    Bit32 offset;
    Bit32 blocks;
    double elapsed;

    if( argc < 3 )
    {
        fprintf( stderr, "usage: %s block_offset block_count\n", program );
        return 1;
    }
    blocks = strtoul( blocks_string, NULL, 10 );
    if( blocks > BIT16_MAX )
    {
        printf( "sorry, %lu is bigger than maximum block count of %u\n"
            "you might want to consider using 0 which means 'until end of disk'\n",
            blocks, BIT16_MAX
        );
        return 2;
    }
    offset = strtoul( offset_string, NULL, 10 );
    printf( "writing same data to %u blocks starting at offset %lu\n", blocks, offset );

    if( aspi_find_disks( &my_srb, MAX_SCSI_DEVICES_PER_BUS, MAX_SCSI_BUSES ) )
        return 3;

    for( count=0; count<aspi_disk_count; count++ )
    {
        printf( "disk at SCSI bus/ID/LUN %d/%d/%d\n",
            aspi_all_disks[count].HostAdaptorNumber,
            aspi_all_disks[count].TargetId,
            aspi_all_disks[count].Lun
        );

        // try writing same data to many logical blocks
        //
        // assumes ASPI driver has been patched to accept physical memory
        // addresses instead of realmode segment:offset addresses
        //
        memset( &my_data, 0x67, BLOCK_SIZE );
        memset( &my_srb, 0, sizeof(AspiSrb) );
        my_scsi_io->Command = AspiCommandExecuteScsiIo;
        my_scsi_io->HostAdaptorNumber = aspi_all_disks[count].HostAdaptorNumber;
        my_scsi_io->RequestFlags.EnablePosting = FALSE;
        my_scsi_io->RequestFlags.EnableResidualByteLengthReporting = TRUE;
        my_scsi_io->RequestFlags.TransferDirection = AspiXferDirectionHostToScsiTarget;
        my_scsi_io->RequestFlags.Reserved = 0;
        my_scsi_io->TargetId = aspi_all_disks[count].TargetId;;
        my_scsi_io->Lun = aspi_all_disks[count].Lun;
        my_scsi_io->DataAllocationLength = BLOCK_SIZE;
        my_scsi_io->SenseAllocationLength = sizeof( my_scsi_io->CdbAndSense.Sense );
        my_scsi_io->DataBufferPointer = real_to_physical( (Bit32)my_data );
        my_scsi_io->SrbLinkPointer = 0;         // we won't be linking
        my_scsi_io->PostRoutineAddress = 0;     // we won't be posting
        my_scsi_io->CdbLength = sizeof( *my_write_same );
        my_write_same->Command = ScsiCommandWriteSame;
        my_write_same->LogicalBlockAddress = htonl( offset );
        my_write_same->NumberOfBlocks = htons( blocks );

        if( !aspi_request_and_wait( &my_srb,
            AspiStatusIgnoreNothing,
            TIMEOUT, &elapsed, FALSE,
            "writing same data to %lu blocks\n",
            blocks
        ) )
        {
            printf( "rate: %g bytes per second\n", BLOCK_SIZE*blocks/elapsed );
        }
    }

    return 0;
}
