/* 1620, Wed 21 Jan 98 */

/* DEBUG flag may be set for my internal playing */

#define noDEBUG

#define J_DEBUG_CLOSING_CA_RTT
#define noJ_DEBUG_RETRANS_TIMEOUT
#define noJ_DEBUG_DUP_ACK
#define noJ_DEBUG_RETRANS_TIMEOUT_KBWAIT
#define noJ_DEBUG_WINDOW_OVERFLOW
#define noJ_DEBUG_TCP_SEND_STARTDATA
#define noJ_DEBUG_READ_WINDOW_UPDATE
#define noJ_DEBUG_TX         /* just dump TX seqnum and acknum */
#define noJ_DEBUG_RX         /* just dump RX seqnum and acknum */

#define noN_DEBUG_CA         /* Congestion avoidance */
#define noN_DEBUG_RTT        /* Round Trip Time estimation */
#define noN_DEBUG_SEND       /* tcp_send() */
#define noN_DEBUG_RECV       /* tcp_ProcessData() */
#define noN_DEBUG_UPDOWN     /* tcp state machine */
#define noN_DEBUG_SOON       /* sendsoon() */

#define noJ_TEST_MISORDERING /* disable queueing of misordered packets */

#define noN_TEST_SEQ_WRAP    /* Check seq nbr wrap-around works properly */
#define noN_TEST_RX_LOSS     /* Drop packets while receiving */
#define noN_TEST_TX_LOSS     /* Drop packets while sending */

/*
 * PCTCP - the true worker of Waterloo TCP
 *       - Contains all opens, closes, major read/write routines
 *         and basic IP handler for incomming packets
 *       - NOTE: much of the TCP/UDP/IP layering is done at the data
 *         structure level, not in separate routines or tasks
 */

#include "..\wattcp\copyrigh.h"
#include <time.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <conio.h>
#include <string.h>
#include <mem.h>
#include <dos.h>
#include <values.h>

#include "..\wattcp\wattcp.h"
#include "..\wattcp\elib.h"
#include "..\types\types.h"
#include "..\dpmi32\phys.h"

/*
 * Local IP address
 */
longword my_ip_addr = 0;  /* For external references */
longword sin_mask = 0xfffffe00;
longword sin_gate = 0x0;


/*
 * IP identification numbers
 */

static int ip_id = 0;                   /* Packet number */
static int next_tcp_port = 1024;        /* Autoincremented */
static int next_udp_port = 1024;
static tcp_Socket *tcp_allsocs = NULL;
static udp_Socket *udp_allsocs = NULL;

/* Timer definitions */
#define RETRAN_STRAT_TIME     7   /* ticks Interval check socket states */
#define SOON_TIMEOUT          4   /* ticks NOT USED ANYMORE!!!  Don't send this close to rtt timeout */
#define tcp_TIMEWT_TIMEOUT    15  /* secs  2 * maximum segment lifetime */
#define tcp_LASTACK_TIMEOUT   0   /* secs  spec says infinite */
#define tcp_CLOSING_TIMEOUT   0   /* secs  spec says infinite */
#define tcp_FINWT1_TIMEOUT    0   /* secs  spec says infinite */
#define tcp_FINWT2_TIMEOUT    675 /* secs  spec says infinite, BSD says 675 */
#define tcp_OPEN_TIMEOUT      75  /* secs  Timeout for opens */
#define tcp_TIMEOUT           13  /* secs  Timeout during a connection */

#define MINPKTRECV  1    // Min buffers free to allow TCP to queue out-of-orders

#define MAXPKTSEND  150  // Max pkts sent per tcp_send() call
//#define MAXPKTSEND  4    // Max pkts sent per tcp_send() call
   // My pkt driver mangled packets if I sent too many
   // of them back to back!            Nevil, 9 Jan 97
   //
   // since we put packets on a global low-level queue when
   // the packet driver tells us they have arrived, they are
   // not dequeued by TCP until application calls tcp_tick(),
   // are we going to find that we lose data even though
   // advertized window is large, because global buffers
   // got full?    if so, maybe we should decrease MAXPKTSEND
   // so TCP doesn't think there is congestion?   nah, prior
   // restraint is less efficient than runtime detection...
   //                                   Joel, 17 Jan 97

word debug_on = 0;

#if (defined(N_TEST_RX_LOSS)) || (defined(N_TEST_TX_LOSS))
extern Bit16 hwt_last_count;  /* In intel\timestmp.cpp */
#endif

#if defined(N_TEST_RX_LOSS) || defined(N_TEST_TX_LOSS)
byte loss_array[] = {1,1,1,1,1,1,1,1, 2,2,2, 3};

int rx_loss_interval = 17;
static int rx_loss_int_ct = 1;
static int rx_loss_nbr_ct = 0;
static Bit32 rx_packet_count = 0;

int tx_loss_interval = 29;
static int tx_loss_int_ct = 1;
static int tx_loss_nbr_ct = 0;
static Bit32 tx_packet_count = 0;
#endif

#if defined(N_DEBUG_RECV) || defined(N_DEBUG_SEND)
Bit32 lastack = 0;
Bit32 lastack_count = 0;
#define DBG_MAX_RTO  150  /* Seconds */
#endif

static void udp_handler(in_Header *ip);
static longword udp_write(udp_Socket *s, byte *datap, longword len);
static longword udp_read(udp_Socket *s, byte *datap, longword maxlen);
static void tcp_Retransmitter(void);


#define TCP_LOCAL 0x4000

/* Statics */
static word tcp_ProcessData(tcp_Socket *s, int *acked, int *ok_to_free,
   in_Header *ip, tcp_Header *tp, word len);
static word tcp_ProcessOptions(tcp_Socket *s, tcp_Header *tp, word len);

static char *mono;
static char *colour;
#if defined(__FLAT__)
static Bit32 mono_handle, colour_handle;
#endif

static initialized = 0;
static void (*system_yield)() = NULL;
extern int multihomes;
extern word _pktipofs;
void (*_dbugxmit)(sock_type *s, in_Header *ih, tcp_Header *ph, char *line)
   = NULL;
void (*_dbugrecv)(sock_type *s, in_Header *ih, tcp_Header *ph, char *line)
   = NULL;
void (*wattcpd)(void) = NULL;

char *_hostname = "012345678901234567890123456789012345678901234567890";

word _mss = ETH_MSS;

char *_wattcp = WATTCP_C;

static int tcp_handler(in_Header *ip, int *rdata_avail);
static void udp_handler(in_Header *ip);
extern void icmp_handler(in_Header *ip);

static void tcp_unthread(tcp_Socket *ds);
static void tcp_abort(tcp_Socket *s);
void tcp_sendsoon(tcp_Socket *s);
static void tcp_send(tcp_Socket *s, int type, char *from);
#   define SEND_NORMAL            0  /* tcp_send() type values */
#   define SEND_FAST_RETRANSMIT   1
#   define SEND_IMMEDIATE_ACK     2
#   define SEND_IF_HAVE_DATA      3
static void tcp_rst(in_Header *his_ip, tcp_Header *oldtcpp);
static void udp_close(udp_Socket *ds);
static void tcp_close(tcp_Socket *s);

#define MAX_MISORDERED  1000
static struct pseg *misordered_all;
static struct segq misordered_free;




#undef min

/* return the smaller of 2 longwords */
static longword _fastcall min( longword left, longword right )
{
    if( left < right )
        return left;

    return right;
}

/* return the smaller of 2 pointers */
static byte *_fastcall ptr_min( void *left, void *right )
{
    if( left < right )
        return left;

    return right;
}

/* Compare two sequence numbers, allowing for wrap-around */
static int seqcmp(Bit32 seq, Bit32 ack)
{
   if (seq == ack) return 0;
   if (seq < ack)
      return ack-seq > 0x0fffffff ? +1 : -1;
   else  /* seq > ack */
      return seq-ack <= 0x0fffffff ? +1 : -1;
   }

/* Put a pseg at the tail of a segq without regard to its seq/len
 *
 * not used by anybody so far
 *
 * Does not actually copy the data into the buffer
 *
 * No need to disable interrupts while running, since none
 * of the interrupt handlers call this code; only the
 * single-threaded foreground code does
 */
void seg_enqueue_tail(struct segq *bq, struct pseg *np)
{
   if (bq->len == 0) bq->head = np;
   else bq->tail->next = np;
   bq->tail = np;
   np->next = NULL;
   ++bq->len;
   }

/* Put a pseg at the head of a segq without regard to its seq/len
 *
 * Used for the free list and for returning a sampled
 *
 * Does not actually copy the data into the buffer
 *
 * No need to disable interrupts while running, since none
 * of the interrupt handlers call this code; only the
 * single-threaded foreground code does
 */
void seg_enqueue_head(struct segq *bq, struct pseg *np)
{
   if (bq->len == 0) {
      bq->tail = np;
      np->next = NULL;
      }
   else
      np->next = bq->head;
   bq->head = np;
   ++bq->len;
   }

/* Take the head pbuf from a queue
 *
 * No need to disable interrupts while running, since none
 * of the interrupt handlers call this code; only the
 * single-threaded foreground code does.
 */
struct pseg *seg_dequeue_head(struct segq *bq)
{
   struct pseg *rb;

   if (bq->len == 0) rb = NULL;
   else {
      rb = bq->head;
      bq->head = rb->next;
      --bq->len;
      }
   return rb;
   }

/* Put a pseg into a segq, maintained in ascending seq order,
 * possibly merging multiple pseg's on the segq together
 *
 * Does not actually copy the data into the buffer, although
 * someday it might be able to do so.
 *
 * No need to disable interrupts while running, since none
 * of the interrupt handlers call this code; only the
 * single-threaded foreground code does
 */
void seg_enqueue_merge( struct segq *bq, struct pseg *np )
{
   struct pseg *right, *left;

   // stupid special case: zero length
   //
   if (np->len == 0) {
      // free caller's segment
      //
      seg_enqueue_head( &misordered_free, np );

      return;
      }

   // special case for empty queue, because head and tail
   // are not guaranteed to be NULL then
   //
   if (bq->len == 0) {
      // we will be the only segment on queue
      //
      bq->head = bq->tail = np;
      np->next = NULL;
      ++bq->len;

      return;
      }

   // find the 2 segments that left edge of new packet fits between
   //
   for( left = NULL, right = bq->head;
        right && seqcmp(np->seq,right->seq)>0;
        left = right, right = right->next )
      ;

   // now make the caller's segment part of the list
   //

   // if previous segment wasn't the head of list
   // (there is a left segment) and right edge of left segment
   // overlaps left edge of caller's segment...
   //
   if( left && seqcmp(left->seq + left->len, np->seq)>=0 ) {

      // move right edge of left segment to include caller's segment
      //
      // don't worry, when somebody or the union is straddling the
      // 4 GB boundary, this math still gives the correct answer
      // because a large negative is the same as a small positive:
      //
      // (3.99G + 0.02G) - 3.95G = 0.01G - 3.95G = -3.94G = +0.06G
      //
      left->len = np->seq + np->len - left->seq;

      // free caller's segment
      //
      seg_enqueue_head( &misordered_free, np );

      }
   // otherwise caller's segment's left edge not overlapping left segment
   else {

      // insert caller's segment between left and right segments
      //
      np->next = right;
      if (left)
         left->next = np;
      else
         bq->head = np;
      if (!right)
         bq->tail = np;
      bq->len++;

      // point to it as left segment
      //
      left = np;

      }

   // keep enlarging left segment until right edge doesn't overlap
   // the segment to its right
   //

   // while there is a right segment and right edge of left segment
   // overlaps left edge of right segment...
   //
   while( right && seqcmp(left->seq + left->len, right->seq)>=0 ) {

      // move right edge of left segment to include right segment
      //
      // see comments above on why the math works even with 4G straddling
      //
      left->len = right->seq + right->len - left->seq;

      // free the right segment
      //
      left->next = right->next;
      seg_enqueue_head( &misordered_free, right );
      bq->len--;

      // the segment one further right is now "right"
      //
      right = left->next;

      }
   }

/*
 * sock_yield - Enable user defined yield function
 */
void sock_yield(tcp_Socket *s, void (*fn)())
{
   if (s) s->usr_yield = fn;
   else system_yield = fn;
   }

/* tells the user how many seconds StateTIMEWT would/did last
 */
longword sock_get_2msl_timeout( tcp_Socket *s )
{
   return tcp_TIMEWT_TIMEOUT;
   }

/*
 * sock_mode - Set binary or ascii - affects sock_gets, sock_dataready
 */
word sock_mode(sock_type *s, word mode)
{
   return s->tcp.sock_mode = (s->tcp.sock_mode & 0xfffc) | mode;
   }

/*
 * sock_ttl - Set time to live
 *            for all packets sent in the future with this socket
 *
 * returns TTL of packets sent in the past
 */
byte sock_ttl(sock_type *s, byte ttl)
{
   byte old_ttl = s->tcp.ttl;
   s->tcp.ttl = ttl;
   return old_ttl;
   }

/*
 * sock_tunnel - Set IP address to tunnel through (in host byte order)
 *               for all packets sent in the future with this socket
 *
 *               0.0.0.0 disables tunneling
 *
 * Returns IP address that past packets were tunneled through
 */
longword sock_tunnel(sock_type *s, longword tunneladdr)
{
   longword old_tunneladdr = s->tcp.tunneladdr;
   s->tcp.tunneladdr = tunneladdr;
   return old_tunneladdr;
   }

/*
 *  IP user level timer stuff
 *  void ip_timer_init(void *s, int delayseconds)
 *  int  ip_timer_expired(void *s)
 *      - 0 if not expired
 */
#define MAXTICKS 0x1800b0L  /* BIOS ticks per day */

void ip_timer_init(udp_Socket *s, int delayseconds)
{
   s->usertimer = delayseconds ? set_timeout(delayseconds) : 0;
   }

int ip_timer_expired(udp_Socket *s)
{
   return s->usertimer ? chk_timeout(s->usertimer)
      : 0;  /* Can't expire */
   }

/* Only used by lpr.c and lpq.c to generate a random
 * 8 bit value, so just get some bits from set_ttimeout
 */
longword MsecClock()
{
   return set_ttimeout(0);
   }

static long make_timeout(word timeout)
{
   return timeout ? set_timeout(timeout) : 0;
   }

/*
 * check_timeout - Test agains timeout clock - account for overflow
 */
static int check_timeout(unsigned long timeout)
{
   return timeout ? chk_timeout(timeout) : 0;
   }

/*
 * Look for bugs
 */
tcp_checkfor(tcp_Socket *t)
{
   tcp_Socket *p;
   for (p = tcp_allsocs ; p ; p = p->next)
      if (p == t) return 1;
   return 0;
}

/*
 * Shut down the card and all services
 */
void tcp_shutdown()
{
   while (tcp_allsocs)
      tcp_abort(tcp_allsocs);
   _eth_release();

#if defined(__FLAT__)
   /* Unmap physical memory for the display
    *
    * WARNING: must do this in reverse of the order that
    * we mapped originally, or we will be making the wrong
    * set of pages invalid when we decrease DS limit
    */
   unmap_physical(4000, mono_handle);
   unmap_physical(4000, colour_handle);
#endif

   initialized = 0;
   }

/*
 * tcp_Init - Initialize the tcp implementation
 *          - May be called more than once without hurting
 */
void tcp_init()
{
   extern int _arp_last_gateway;
   extern int _last_nameserver;
   unsigned amount = sizeof( struct pseg ) * MAX_MISORDERED;

   if (!initialized) {
#if defined(__FLAT__)
      /* Map physical memory to flat memory for the display */
      colour = map_physical(0xB8000, 4000, &colour_handle);
      mono = map_physical(0xB0000, 4000, &mono_handle);
#else
      mono = (char *)0xb0000000L;
      colour = (char *)0xb8000000L;
#endif
      /* Initialize Ethernet interface */
      initialized = 1;
      _eth_init();

      /* Reset the various tables */
      _arp_last_gateway = 0;  /* Reset the gateway table */
      _last_nameserver = 0;   /* Reset the nameserver table */
      _last_cookie = 0;       /* Eat all remaining crumbs */
      *_hostname = 0;         /* Reset the host's name */

      _eth_free(NULL);

      next_udp_port = next_tcp_port = 1024 + (set_timeout(0) & 0x1ff);

      /* allocate memory for global misordered-segment free list */
      misordered_all = malloc( amount );
      if (!misordered_all) {
         fprintf (stderr, "cannot allocate %u bytes for %u misordered segment pointers\naborting\n", amount, MAX_MISORDERED );
         exit( 4 );
         }

      /* link all the blocks together on the free list */
      for( misordered_free.len=0; misordered_free.len<MAX_MISORDERED; ) {
         seg_enqueue_head( &misordered_free, misordered_all+misordered_free.len );
         }
      }
   }

/* Turn off compiler warning about unused parameters s and size */
#pragma warn -par

/*
 * Checks for bugs when compiling in large model C compiler
 *
 * Borland C uses a 4K stack by default.  In all memory models the
 * stack grows down toward the heap.
 *
 * If you accidentally place tcp_Socket onto the stack (like by making
 * it an automatic variable), then you will have already used up that
 * whole 4K and then some!
 *
 * In large model, this will mess up the data space in a major way
 * because the stack starts at SS:_stklen, or SS:1000, so you will
 * wrap the SP pointer back around to FFFE and start writing over
 * the far heap.  Yuck.
 *
 * In small model it usually doesn't kill your application because
 * you would have to be down to your last 4K of memory and this is
 * not as common.
 *
 * The solutions: declare your sockets as static, or put them on the
 * heap, or bump up your stack size by using the global special variable:
 *
 * unsigned _stklen = 16536;    // set stack to 16 k
 */

static void largecheck(void *s, int size)
{
   }

/* Restore compiler warning about unused parameters */
#pragma warn .par

/*
 * findfreeport - Return unused local port
 *              - oldport = 0:normal port, 1:special port (513-1023)
 *              - We need not be this picky, but it doesn't hurt
 */
static word findfreeport(word oldport)
{
   word temp;
   tcp_Socket *s;

   if (oldport > 0 && oldport < 0xffff)
      return oldport;

   if (oldport == 0) oldport = 1025;
   else oldport = 513;

   for (temp = oldport ; temp < oldport + 510 ; ++temp) {
      if ((s = (tcp_Socket*)udp_allsocs) != NULL) {
         while (s->next && s->myport != temp)
            s = (tcp_Socket*)s->next;
         if (s->myport == temp) continue;
         }
      if ((s = tcp_allsocs) != NULL) {
         while (s->next && s->myport != temp)
            s = s->next;
         if (s->myport == temp) continue;
         }
      break;
      }
   return temp;
   }

/* socket, localport, destaddress */
int udp_open(udp_Socket *s,
   word lport, longword ina, word port, dataprocref rdataHandler)
{
   // preserve user-provided buffering if it is in the socket
   //
   longword    maxrdatalen = s->maxrdatalen;
   byte       *rdata = s->rdata;

   udp_close(s);
   largecheck(s, sizeof(udp_Socket));

   memset(s, 0, sizeof(udp_Socket));
   s->ip_type = IPPROTO_UDP;

   if( !sock_set_rx_buf( (tcp_Socket *)s, rdata, maxrdatalen ) )
      return 0;

   s->ttl = in_DefaultTTL;

   lport = findfreeport(lport);

   s->myport = lport;
   s->myaddr = my_ip_addr;
   s->tunneladdr = 0;

   /* Check for broadcast */
   if ((long)(ina) == -1 || !ina)
      memset(s->hisethaddr, 0xff, sizeof(eth_address));
   else if (! _arp_resolve(ina, (eth_address *) &s->hisethaddr[0], 0))
      return 0;

   s->hisaddr = ina;
   s->hisport = port;
   s->rdataHandler = rdataHandler;
   s->usr_yield = system_yield;
   s->safetysig = SAFETYUDP;
   s->next = udp_allsocs;
   udp_allsocs = s;
   return 1;
   }

/*
 * Initialize socket data structures for either tcp_open()
 * or tcp_connect(), possibly choosing the local port
 *
 * Returns 0 on error
 */
int tcp_open(tcp_Socket *s, word lport)
{
   // fail if look for sockets in StateTIMEWT with same local port

   largecheck(s, sizeof(tcp_Socket));  /* Stack space warnings */
   tcp_unthread(s);  /* Just in case not totally closed */

   memset(s, 0, sizeof(tcp_Socket));
   s->ip_type = IPPROTO_TCP;

   /* start using default RX and TX buffering */
   sock_set_rx_buf( s, 0, 0 );
   sock_set_tx_buf( s, 0, 0 );

   /* didn't get WSCALE option from other side yet */
   s->rx_wscale = BIT8_MAX;

   s->ttl = in_DefaultTTL;
   s->mss = _mss;
   s->state = tcp_StateOPEN;

   s->sA = 0;  /* Nevil, 30 Dec 97 */
   s->sD = 3*18 << 2;  /* D = 3 seconds */
   s->AD_ok = 0;  /* Waiting for first RTT */
   s->ssthresh = 65535 / s->mss;
   s->rto = 0 + 2 * 3*18;  /* 2, not 4!  Stevens, p 305 */
   s->cwnd = 1;  /* Slow start */
   s->wwnd = 0;
   s->wwnd2 = 0;
#ifdef N_DEBUG_RTT
   printf("tcp_open(): vj(%u %u %u rto=%u)\n",
      s->AD_ok, s->sA, s->sD, s->rto);
#endif

   lport = findfreeport(lport);  /* Get a nonzero port val */
   s->myaddr = my_ip_addr;
   s->myport = lport;
   s->tunneladdr = 0;           /* no tunneling by default */

   s->seqnum = htonl(set_timeout(0));  /* Make up a number */
#ifdef N_TEST_SEQ_WRAP
   s->seqnum |= 0xffffc000 | hwt_last_count;
#endif
   s->dbg_tx_seqbase = s->seqnum;
   s->usr_yield = system_yield;

   s->safetysig = SAFETYTCP;       /* Insert into chain */
   s->next = tcp_allsocs;
   tcp_allsocs = s;

   return 1;
   }

/*
 * Actively open a TCP connection to a particular host/port
 *      - 0 on error
 */
int tcp_connect(tcp_Socket *s,
   longword ina, word port, dataprocref rdataHandler)
{
   tcp_Socket *s2;

   if (s->state!=tcp_StateOPEN || s->safetysig!=SAFETYTCP) return 0;
   if (ina - my_ip_addr <= multihomes) return 0;

   // fail if any socket is in StateTIMEWT with same 4-tuple
   for (s2=tcp_allsocs; s2; s2=s2->next) {
      if (s2->state == tcp_StateTIMEWT &&
            s2->myaddr == s->myaddr &&
            s2->myport == s->myport &&
            s2->hisaddr == ina &&
            s2->hisport == port) {
         return 2;
         }
      }

   if (!_arp_resolve(ina, (eth_address *) &s->hisethaddr[0], 0))
      return 0;

   s->hisaddr = ina;
   s->hisport = port;
   s->rdataHandler = rdataHandler;
   s->state = tcp_StateSYNSENT;
   s->abort_timeout = set_timeout(tcp_OPEN_TIMEOUT);
   s->flags = tcp_FlagSYN;
   tcp_send(s, SEND_NORMAL, "tcp_connect()");
   return 1;
   }

/*
 * Passive open: listen for a connection from a particular host/port
 */
int tcp_listen(tcp_Socket *s,
   longword ina, word port, dataprocref rdataHandler, word timeout)
{
   if (s->state!=tcp_StateOPEN || s->safetysig!=SAFETYTCP) return 0;

   s->hisport = port;
   s->hisaddr = ina;
   s->rdataHandler = rdataHandler;
   s->state = tcp_StateLISTEN;
   if (timeout == 0) s->abort_timeout = 0; /* Forever ... */
   else s->abort_timeout = set_timeout(timeout);

   return 1;
   }


static void udp_close(udp_Socket *ds)
{
   udp_Socket *s, **sp;

   sp = &udp_allsocs;
   for (;;) {
      s = *sp;
      if (s == ds) {
         *sp = s->next;
         break;
         }
      if (!s) break;
      if (!s->err_msg) s->err_msg = "UDP Close called";
      sp = &s->next;
      }
   }

/*
 * Send a FIN on a particular port -- only works if it is open
 *   Must still allow receives
 */
static void tcp_close(tcp_Socket *s)
{
   if (s->ip_type != IPPROTO_TCP)
      return;
   if (s->state == tcp_StateESTAB ||
         s->state == tcp_StateESTCL ||
         s->state == tcp_StateSYNREC) {

      if (s->sdatalen) {  /* Must first flush all data */
         s->flags |= tcp_FlagPUSH | tcp_FlagACK;
         if (s->state < tcp_StateESTCL) {
             s->state = tcp_StateESTCL;
#ifdef N_DEBUG_SOON
             printf("sendsoon from tcp_close(ESTCL)\n");
#endif
             tcp_sendsoon(s);
             }
         }
      else { /* Really closing */
#ifdef J_DEBUG_CLOSING_CA_RTT
         printf("tcp_close(active): sA=%u sD=%u rto=%u cwnd=%u\n",
            s->sA, s->sD, s->rto, s->cwnd);
#endif
         s->flags = tcp_FlagACK | tcp_FlagFIN;
         if (!s->err_msg)
            s->err_msg = "Connection actively closed normally";
         s->state = tcp_StateFINWT1;
         s->abort_timeout = tcp_FINWT1_TIMEOUT ? set_timeout(tcp_FINWT1_TIMEOUT) : 0;
         tcp_send(s, SEND_NORMAL, "tcp_close(FINWT1)");
         }
      }
   else if (s->state == tcp_StateCLOSWT ||
         s->state == tcp_StateCLOSWTCL) {
      if (s->sdatalen) {  /* Must first flush all data */
         s->flags |= tcp_FlagPUSH | tcp_FlagACK;
         if (s->state != tcp_StateCLOSWTCL) {
             s->state = tcp_StateCLOSWTCL;
#ifdef N_DEBUG_SOON
             printf("sendsoon from tcp_close(CLOSWTCL)\n");
#endif
             tcp_sendsoon(s);
             }
         }
      else { /* Really closing */
         /* Need to ack the fin and get on with it */
#ifdef J_DEBUG_CLOSING_CA_RTT
         printf("tcp_close(passive): sA=%u sD=%u rto=%u cwnd=%u\n",
            s->sA, s->sD, s->rto, s->cwnd);
#endif
         s->flags |= tcp_FlagFIN;
         if (!s->err_msg)
            s->err_msg = "Connection passively closed normally";
         s->state = tcp_StateLASTACK;
         s->abort_timeout = tcp_LASTACK_TIMEOUT ? set_timeout(tcp_LASTACK_TIMEOUT) : 0;
         tcp_send(s, SEND_NORMAL, "tcp_close(LASTACK)");
         }
      }
   }

/*
 * Abort a tcp connection
 */
static void tcp_abort(tcp_Socket *s)
{
   if (!s->err_msg) s->err_msg = "TCP_ABORT";
   if (s->state != tcp_StateLISTEN &&
         s->state != tcp_StateCLOSED &&
         s->state != tcp_StateTIMEWT) {
#ifdef J_DEBUG_CLOSING_CA_RTT
      printf("tcp_abort(resetting): sA=%u sD=%u rto=%u cwnd=%u\n",
         s->sA, s->sD, s->rto, s->cwnd);
#endif
      s->flags = tcp_FlagRST | tcp_FlagACK ;
      tcp_send(s, SEND_NORMAL, "tcp_abort()");
      }
   s->sdatalen = 0;
   s->sdataoffset = 0;
   s->ip_type = 0;
   s->state = tcp_StateCLOSED;
   if (s->rdataHandler) s->rdataHandler(s, 0, -1, 0, 0);
   tcp_unthread(s);
   }

void sock_abort(sock_type *s)
{
   if (s->udp.ip_type == IPPROTO_TCP)
      tcp_abort(&s->tcp);
   else if (s->udp.ip_type == IPPROTO_UDP)
      udp_close(&s->udp);
   /* Do nothing when socket is closed */
   }

/*
 * tcp_sendsoon - Schedule a transmission pretty soon
 */
void tcp_sendsoon(tcp_Socket *s)
{
   if (s->ip_type == IPPROTO_TCP)
      s->soon_req = 1;
   }

static void set_ssthresh(tcp_Socket *s)
{
   longword w = s->window / s->mss;  /* His window size */
   s->ssthresh = (s->cwnd < w ? s->cwnd : w)/2;
   if (s->ssthresh < 2) s->ssthresh = 2;
   s->cwnd = s->ssthresh;
   s->wwnd = 0;
   s->wwnd2 = 0;
   }

/*
 * Retransmitter - called periodically to perform tcp retransmissions
 */
static longword retran_strat = 0L; /* Timer for retran strategy */
static void tcp_Retransmitter(void)
{
   tcp_Socket *s;
   int sent;

   /* Only do this once per RETRAN_STRAT_TIME ticks */
   if (!chk_timeout(retran_strat)) return;
   retran_strat = set_ttimeout(RETRAN_STRAT_TIME);

   for (s = tcp_allsocs; s; s = s->next) {
      sent = 0;
      if ((s->sdatalen ||  /* This socket has sent data */
            s->state == tcp_StateSYNSENT) &&    /* trying to open connection */
            chk_timeout(s->retrans_timeout)) {  /* Tx timeout */

#ifdef J_DEBUG_RETRANS_TIMEOUT
         printf("before retransmit timeout: ssthresh=%u, cwnd=%u, rto=%u, sA=%u, sD=%u\n",
            s->ssthresh, s->cwnd, s->rto, s->sA, s->sD );
#ifdef J_DEBUG_RETRANS_TIMEOUT_KBWAIT
         printf("press a key to continue:\n" );
         getch();
#endif
#endif
#ifdef DEBUG
         if (debug_on >1) printf(
            "regular retran TO set unacked back to 0 from %lu\n",
            s->unacked);
#endif DEBUG
         s->rto *= 2;  /* Exponential backoff (Stevens p. 299) */
         if (s->rto > 64*18)
            s->rto = 64*18;
         set_ssthresh(s);
         s->cwnd = 2;  /* Slow start */
         s->unacked = 0;  /* Retransmit from last ACKed seqnbr */
#ifdef N_DEBUG_CA
         printf("Retransmitter timeout: ssthresh=%u, cwnd=%u\n", s->ssthresh, s->cwnd);
#endif
#if defined(N_DEBUG_RECV) || defined(N_DEBUG_SEND)
         if (s->rto > DBG_MAX_RTO*18) {  /* seconds */
            s->err_msg = "rto > DBG_MAX_RTO";
            exit(3);
            }
#endif
         if (s->sdatalen)
            s->flags |= tcp_FlagPUSH | tcp_FlagACK;
         tcp_send(s, SEND_NORMAL, "tcp_Retransmit(timeout)");
         /* per Karn's Algorithm, see Stevens volume 1, page 301,
          * don't calculate RTT using this packet, because the original
          * packet might (somehow) get to the receiver, and the next ACK
          * we see could have been transmitted before they got this one
          *
          * Stevens volume 2, pages 842-844 show that the 4.4BSD-lite
          * code of 1993 only clears the timer, rather than remembering
          * the right edge of the first segment retransmitted to know
          * when we have gotten the ACK for it and it is OK to start
          * timing segments, because the beginning of slow-start naturally
          * restricts us from having more than one segment outstanding
          * anyways.
          *
          * Unfortunately, the use of the phrase "retransmitted data"
          * in the statements "don't calculate a new RTO until
          * an acknowledgement is received for a segment that was not
          * retransmitted" in volume 1, and "the only segments that are ...
          * used to update the RTO are those that are not retransmitted"
          * in volume 2, can too-easily be misconstrued to mean "all bytes
          * that were unacknowledged before the timeout.
          *
          * The hidden assumption is that the TCP timeout mechanism is
          * for a single segment, not a range of bytes.  It protects against
          * gross errors like link cut/restoration or router/receiver tail
          * drop, where a large set of packets just didn't get there AND
          * NEVER WILL.  So you only want to back off exponentially if no
          * ACK's come in at all, not all day until full retransmission of
          * all data that used to be outstanding in a many-megabyte window
          * is achieved.
          *
          * The previous implementation (which remembered the sequence
          * number for all data we have outstanding now, and refused to
          * time any packets until something beyond them got sent) was
          * doubly bad because it not only kept us from using a reasonable
          * estimate of RTT as the RTO for timeouts in the (large) window;
          * each time a timeout occurred in the window it backed off further!
          */
         s->tx_time = 0;
         sent = 1;
         }
      if (!sent && s->soon_req) {
         tcp_send(s, SEND_NORMAL, "tcp_Retransmit(soon)");
         }

      /* Handle inactive tcp timeouts */
      if (sock_inactive && s->inactive_to) {
         if (chk_timeout(s->inactive_to)) {
            /* This baby has timed out */
            s->err_msg = "Connection timed out - no activity";
            sock_close((sock_type *) s);
            }
         }

      if (s->abort_timeout && chk_timeout(s->abort_timeout) &&
            s->state != tcp_StateESTAB && s->state != tcp_StateESTCL) {
         if (s->state != tcp_StateTIMEWT)
            s->err_msg = "Timeout, aborting";
         tcp_abort(s);
         break;
         }
      }

   /* Do our various daemons */
   if (wattcpd) (*wattcpd)();
   }


/*
 * Unthread a socket from the tcp socket list, if it's there
 */
static void tcp_unthread(tcp_Socket *ds)
{
   tcp_Socket *s, **sp;

   if (!ds->rdatalen || (ds->state > tcp_StateESTCL))
      ds->ip_type = 0;  /* Fail IO */
   ds->state = tcp_StateCLOSED;   /* tcp_tick needs this */
   sp = &tcp_allsocs;
   for (;;) {
      s = *sp;
      if (s == ds) {
         *sp = s->next;
         continue;  /* Unthread multiple copies if necessary */
         }
      if (!s) break;
      sp = &s->next;
      }
   }

/* Turn off compiler warning about incorrect assignment
 * since I don't want to duplicate this code inside and
 * outside the while () statement */
#pragma warn -pia

/*
 * tcp_tick - Called periodically by user application
 *          - Returns 0 when s closes (IPPROTO_TCP or IPPROTO_UDP while open)
 *          - Called with socket parameter or NULL
 */
int tcp_tick(sock_type *s)
{
   in_Header *ip;
   int ok_to_free, rdata_avail;
   word packettype;

   /* Finish off dead sockets */
   if (s) {
      if ((s->tcp.ip_type == IPPROTO_TCP) &&
            (s->tcp.state == tcp_StateCLOSED) &&
            (s->tcp.rdatalen == 0)) {
         tcp_unthread(& s->tcp);
         s->tcp.ip_type = 0;
         }
      }

   while (ip = (in_Header *)_eth_arrived(&packettype)) {
      ok_to_free = 1;  rdata_avail = 0;
      switch (packettype) {
      case 0x008:  /* 0x0800  IP */
         if (checksum(ip, in_GetHdrlenBytes(ip)) == 0xffff) {
            switch (ip->proto) {
            case IPPROTO_TCP:
               ok_to_free = tcp_handler(ip, &rdata_avail);
               break;
            case IPPROTO_UDP:
               udp_handler(ip);
               break;
            case IPPROTO_ICMP:
               icmp_handler(ip);
               break;
               }
             }
         else  {
            if (debug_on)
               outs("IP Received BAD Checksum \n");
            }
         break;
      case 0x608:  /* 0x0806  ARP */
         _arp_handler((arp_Header *)ip);
         break;
         }
      if (ok_to_free) _eth_free(ip);
      if (rdata_avail) {  /* Our rx buffer is getting full */
         /* Let user program try tcp_read (probably via sock_read) */
         return s ? s->udp.ip_type : 0;
         }
      }

   /* Check for our outstanding packets */
   tcp_Retransmitter();

   return s ? s->udp.ip_type : 0;  /* Nevil, 31 Mar 93 */
   }

/* Restore compiler warning about incorrect assignment */
#pragma warn .pia

void tcp_set_debug_state(int x)
{
   debug_on = x;
   }

/* Returns 1 if connection is established */
int tcp_established(tcp_Socket *s)
{
   return s->state >= tcp_StateESTAB;
   }


static longword udp_write(udp_Socket *s, byte *datap, longword len)
{
   tcp_PseudoHeader ph;
   struct _pkt {
      in_Header  in;
      udp_Header udp;
      int        data;
/*    longword maxsegopt; */
      } *pkt;
   struct _tunnel {
      in_Header  in;
      struct _pkt pkt;
      } *tunnel;
   byte *dp;
   in_Header *inp;
   udp_Header *udpp;

   if (s->tunneladdr) {
      tunnel = (struct _tunnel *)_eth_formatpacket(&s->hisethaddr[0],
         /*0x800*/ 8);
      pkt = &tunnel->pkt;
      if (len > ETH_MAXUDP - sizeof(in_Header))
         len = ETH_MAXUDP - sizeof(in_Header);
      }
   else {
      pkt = (struct _pkt *)_eth_formatpacket(&s->hisethaddr[0],
         /*0x800*/ 8);
      if (len > ETH_MAXUDP)
         len = ETH_MAXUDP;
      }
   dp = (byte *) &pkt->data;
   inp = &pkt->in;
   udpp = &pkt->udp;

   memset(inp, 0, sizeof(in_Header));
   inp->length = htons(sizeof(in_Header) + UDP_LENGTH + len);

   /* UDP header */
   udpp->srcPort = htons(s->myport);
   udpp->dstPort = htons(s->hisport);
   udpp->checksum = 0;
   udpp->length = htons(UDP_LENGTH + len);
   movmem(datap, dp, len);

   /* IP header */
   inp->ver = 4;
   inp->hdrlen = 5;
   inp->tos = 0;
   inp->identification = htons(++ip_id);   /* was post inc */
//    inp->frag = 0;
   inp->ttl = s->ttl;
   inp->proto = IPPROTO_UDP;
   inp->checksum = 0;
   inp->source = htonl(s->myaddr);
   inp->destination = htonl(s->hisaddr);
   inp->checksum = ~checksum(inp, sizeof(in_Header));

   /* Compute udp checksum if desired */
   if (s->sock_mode & UDP_MODE_NOCHK)
      udpp->checksum = 0;
   else {
      ph.src = inp->source;  /* Already INTELled */
      ph.dst = inp->destination;
      ph.mbz = 0;
      ph.protocol = IPPROTO_UDP;
      ph.length = udpp->length;  /* Already INTELled */

      ph.checksum = checksum(&pkt->udp, ntohs(ph.length));
      udpp->checksum =  ~checksum(&ph, sizeof(ph));
      }

   if (_dbugxmit)
      (*_dbugxmit)((sock_type *)s,inp,(tcp_Header *)udpp, "udp_write()");

   if (s->tunneladdr) {
      inp = &tunnel->in;

      inp->ver = 4;
      inp->hdrlen = 5;
      inp->tos = 0;
      inp->identification = htons(++ip_id);  /* Was post inc */
      inp->frags = 0;
      inp->fo = 0;
      inp->ttl = in_DefaultTTL;
      inp->proto = IPPROTO_IP_IN_IP;  /* Please unwrap me! */
      inp->source = htonl(s->myaddr);
      inp->destination = htonl(s->tunneladdr);
      inp->length = htons(sizeof(in_Header) + ntohs(tunnel->pkt.in.length));
      inp->checksum = 0;
      inp->checksum = ~checksum(inp, sizeof(in_Header));
      }

   if (_eth_send(ntohs(inp->length)))
      len = 0;  /* Send failed, tell the caller */

   return len;
   }

/*
 * udp_read - Read data from buffer, does large buffering
 */
static longword udp_read(udp_Socket *s, byte *datap, longword maxlen)
{
   longword x, up_to_end;

   if ((x = s->rdatalen) != 0) {
      if (x > maxlen) x = maxlen;
      if (x > 0) {
         if (datap) {
            up_to_end = s->maxrdatalen - s->rdataoffset;
            if (up_to_end < x) {
               movmem(s->rdata + s->rdataoffset, datap, up_to_end);
               movmem(s->rdata, datap+up_to_end, x-up_to_end);
               }
            else movmem(s->rdata + s->rdataoffset, datap, x);
            }

         s->rdatalen = s->rdatalen > x ? s->rdatalen-x : 0;
         if (s->rdatalen) {  /* Still some data in buffer */
            s->rdataoffset = wrap(s->rdataoffset+x, s->maxrdatalen);
            }
         }
      }
   return x;
   }

void _udp_cancel(in_Header *ip)
{
   int len;
   udp_Header *up;
   udp_Socket *s;

   /* Match to a udp socket */
   len = in_GetHdrlenBytes(ip);
   up = (udp_Header *)((byte *)ip + len);  /* UDP frame pointer */

   /* Demux to active sockets */
   for (s = udp_allsocs; s; s = s->next)
      if (s->hisport != 0 &&
            ntohs(up->dstPort) == s->hisport &&
            ntohs(up->srcPort) == s->myport &&
            ntohl(ip->destination) == s->hisaddr)
         break;
   if (!s) {  /* Demux to passive sockets */
      for (s = udp_allsocs; s; s = s->next)
         if (s->hisport == 0 && ntohs(up->dstPort) == s->myport)
            break;
      }
   if (s) {
      s->rdatalen = 0;
      s->rdataoffset = 0;
      s->ip_type = 0;
      }
   }

void *_tcp_lookup(longword hisip, word hisport, word myport)
{
   tcp_Socket *s;
   for (s = tcp_allsocs; s; s = s->next) {
      if (myport == s->myport &&  /* Always unique under WATTCP */
            hisport == s->hisport &&
            hisip == s->hisaddr)
         return s;
      }
   return NULL;
   }


void _tcp_cancel(in_Header *ip, int code, char *msg, longword dummyip)
/* Called by the ICMP handler (in pcimp.cpp) */
{
   int len;
   tcp_Socket *s;
   tcp_Header *tp;

   len = in_GetHdrlenBytes(ip);  /* Check work */

   tp = (tcp_Header *)((byte *)ip + len);   /* TCP frame pointer */

   /* Demux to active sockets */
   for (s = tcp_allsocs; s; s = s->next) {
      if (ntohs(tp->srcPort) == s->myport &&
            ntohs(tp->dstPort) == s->hisport &&
            ntohl(ip->destination) == s->hisaddr) {
         switch (code) {
         case 1:  /* ICMP 'Destination Unreachable' */
         case 2:  /* ICMP 'Source Quench' */
            set_ssthresh(s);  /* Slow it down */
            break;
         case 5:  /* ICMP 'Redirect' */
            /* Save his NEW network address */
            _arp_resolve(dummyip, &s->hisethaddr, 0);
            break;
            }
         }
      }
   }

/* Joel 20 May 1998
 *
 * This code used to return only the amount of unused buffering
 * beyond the right edge of the out-of-order segments, which
 * eventually caused the sender to be unable to fill in the holes!
 *
 * From Stevens Volume I, page 280-281, section 20.3 :
 *
 * The advertized window indicates how much room is allocated
 * for buffering data beyond that which has been acknowledged.
 *
 * When the left edge of window moves to the right because
 * we acknowledged new data, it is called "closing".
 *
 * When the right edge of window moves to the right because
 * the application has read from the socket, it is called "opening".
 *
 * When the right edge of window moves to the left, possibly
 * because the kernel or application deallocated buffers, it
 * is called "shrinking" and is a VERY BAD thing.
 *
 * Moving the left edge of the window to the left has no name,
 * but would indicate we threw away data that was properly received!
 */
static Bit16 calc_rx_window(tcp_Socket *s)
{
   Bit32 recv_window;

   recv_window = s->maxrdatalen - s->rdatalen;
   if ((s->flags & tcp_FlagSYN) == 0)
      recv_window >>= s->tx_wscale;
   return min( recv_window, 0xFFFF );
   }

// returns the smaller of:
//
// unsent data
// space in receiver's buffer
// space in sender's congestion window
//
static longword sendable_data(tcp_Socket *s)
{
   longword cwnd_bytes = s->cwnd * (longword)s->mss;
   if (cwnd_bytes <= s->unacked)
      return 0;

   return min( s->sdatalen, min( s->window, cwnd_bytes ) ) - s->unacked;
   }

static longword tcp_read(tcp_Socket *s, byte *datap, longword maxlen)
{
   longword x, up_to_end;

   if ((x = s->rdatalen) != 0) {
      if (x > maxlen) x = maxlen;
      if (x > 0) {
         if (datap) {
            up_to_end = s->maxrdatalen - s->rdataoffset;
            if (up_to_end < x) {
               movmem(s->rdata + s->rdataoffset, datap, up_to_end);
               movmem(s->rdata, datap+up_to_end, x-up_to_end);
               }
            else movmem(s->rdata + s->rdataoffset, datap, x);
            }
         s->rdatalen -= x;
         s->rdataoffset = wrap(s->rdataoffset+x, s->maxrdatalen);

   /* In general, should avoid updating the window when it
    * hasn't changed enough to be visible in the TCP window
    * field, even if scaling were implemented -- best way to
    * do that will be to keep a copy of last window we sent,
    * precalculate the window here, and compare them
    *                                     Joel, 17 Jan 97 */

         if (calc_rx_window(s) != s->oldwindow) {
#ifdef J_DEBUG_READ_WINDOW_UPDATE
            printf("tcp_read() updating window soon\n");
#endif
#ifdef N_DEBUG_SOON
            printf("sendsoon from tcp_read()\n");
#endif
            tcp_sendsoon(s);  /* Update the window */
            }
         }
      }
   else if (s->state == tcp_StateCLOSWT)
      tcp_close(s);
   return x;
   }

/*
 * Write data to a connection.
 * Returns number of bytes written, == 0 when connection is not in
 * established state.
 */
static longword tcp_write(tcp_Socket *s, byte *dp, longword len)
{
   longword x, up_to_end, startpos;

   if (s->state != tcp_StateESTAB) len = 0;
   if (len > (x = s->maxsdatalen - s->sdatalen)) len = x;
   if (len > 0) {
      startpos = wrap(s->sdataoffset + s->sdatalen, s->maxsdatalen);
      up_to_end = s->maxsdatalen - startpos;
      if (up_to_end < len) {
         movmem(dp, s->sdata + startpos, up_to_end);
         movmem(dp+up_to_end, s->sdata, len-up_to_end);
         }
      else movmem(dp, s->sdata + startpos, len);

      /* Note: data offset doesn't change when we add to buffer */
      s->sdatalen += len;

      if (s->sock_mode & TCP_LOCAL)
         s->sock_mode &= ~TCP_LOCAL;
      else {
         if (sendable_data(s)) {
            if (s->sock_mode & TCP_MODE_NONAGLE) {
               tcp_send(s, SEND_NORMAL, "tcp_write(NN)");
               }
            else {
               if (s->sdatalen >= s->mss)
                  tcp_send(s, SEND_NORMAL, "tcp_write()");
               else
#ifdef N_DEBUG_SOON
                  printf("sendsoon from tcp_write()\n");
#endif
                  tcp_sendsoon(s);
               }
            }
         }
      }
   return len;
   }

/*
 * Send pending data
 */
static void tcp_Flush(tcp_Socket *s)
{
   if (s->sdatalen) {
      s->flags |= tcp_FlagPUSH;
      tcp_send(s, SEND_NORMAL, "tcp_Flush()");
      }
   }

/*
 * Handler for incoming packets.
 */
static void udp_handler(in_Header *ip)
{
   udp_Header *up;
   tcp_PseudoHeader ph;
   word len;
   byte *dp;
   longword temp;
   udp_Socket *s;

   temp = ntohl(ip->destination);

   // temp = ip number
   //     or 255.255.255.255
   //     or sin_mask.255.255

   if ( (~temp & ~sin_mask) != 0 &&  /* Not a broadcast packet*/
         (temp - my_ip_addr > multihomes  /* Not my address */
            && my_ip_addr) )              /*    and I know my address */
       return;


   len = in_GetHdrlenBytes(ip);
   up = (udp_Header *)((byte *)ip + len);      /* UDP segment pointer */
   len = ntohs(up->length);

   /* Demux to active sockets */
   for (s = udp_allsocs; s; s = s->next) {
      if (s->safetysig != SAFETYUDP) {
         if (debug_on) outs("chain error in udp\r\n");
         }
      if (s->hisport != 0 &&
            ntohs(up->dstPort) == s->myport &&
            ntohs(up->srcPort) == s->hisport &&
            ((ntohl(ip->destination) & sin_mask) == (s->myaddr & sin_mask)) &&
            ntohl(ip->source) == s->hisaddr )
         break;
      }

   if (!s) {
      /* Demux to passive sockets */
      for (s = udp_allsocs; s; s = s->next)
         if (((s->hisaddr == 0) || (s->hisaddr == 0xffffffff))
               && ntohs(up->dstPort) == s->myport) {

            movmem(((eth_Header *)ip - 1)->source,
               s->hisethaddr, sizeof(eth_address));
            s->hisaddr = ntohl(ip->source);
            s->hisport = ntohs(up->srcPort);

            break;
            }
      }
   if (!s) { /* Demux to broadcast sockets */
      for (s = udp_allsocs; s; s = s->next)
         if ((s->hisaddr == 0xffffffff) &&
               (ntohs(up->dstPort) == s->myport))
            break;
      }
   if (!s) {
      if (debug_on > 2) outs("Discarding ...");
      return;
      }
   if (_dbugrecv)
      (*_dbugrecv)((sock_type *)s,ip,(tcp_Header *)up,  "udp_handler()");

   /* These parameters are used for things other than just checksums */
   ph.src = ip->source;  /* Already INTELled */
   ph.dst = ip->destination;
   ph.mbz = 0;
   ph.protocol = IPPROTO_UDP;
   ph.length = up->length;
   if (up->checksum) {
      ph.checksum =  checksum(up, len);
      if (checksum(&ph, sizeof(tcp_PseudoHeader)) != 0xffff)
         return;
      }

   /* Process user data */
   if ((len -= UDP_LENGTH) > 0) {
      dp = (byte *)(up);
      if (s->rdataHandler)
         s->rdataHandler(s, &dp[UDP_LENGTH], len, &ph, up);
      else {
         // just overwrite the beginning of buffer...
         //
         // should this preserve existing data?
         //
         if (len > s->maxrdatalen) len = s->maxrdatalen;
         movmem(&dp[ UDP_LENGTH ], s->rdata, len);
         s->rdatalen = len;
         s->rdataoffset = 0;
         }
      }
   }

static int tcp_handler(in_Header *ip, int *rdata_avail)
/* Returns 1 if finished with packet,
           0 if we've queued the packet in a socket queue
   Nevil, 7 Jan 98 */
{
   tcp_Header *tp;
   tcp_PseudoHeader ph;
   word len;
   byte *dp;
   tcp_Socket *s;
   word flags;
   int M;  /* For estimating rtt (must be signed) */
#ifdef N_DEBUG_RTT
   int M_before;
#endif
   int r;
   longword diff;
   Bit32 pkt_ack;
   Bit32 pkt_seq;
   Bit32 pkt_window;
   int rx_len;  /* Incoming TCP data received */
   int acked, sent, should_send;
   int ok_to_free = 1;

#ifdef N_TEST_RX_LOSS
   if (--rx_loss_int_ct == 0) {
      rx_loss_int_ct = rx_loss_interval + (hwt_last_count >> 8) % 13;
      if (rx_packet_count != 0) {
         rx_loss_nbr_ct =
            loss_array[(hwt_last_count >> 4) % sizeof(loss_array)];
         }
      }
   ++rx_packet_count;
   if (rx_loss_nbr_ct) {
      --rx_loss_nbr_ct;  /* Discard this packet */
      printf("****** TEST: ****** RX packet %lu discarded ******\n",
         rx_packet_count);
      return ok_to_free;
      }
#endif

   if ((longword)(ntohl(ip->destination) - my_ip_addr) > multihomes)
      return ok_to_free;

   len = in_GetHdrlenBytes(ip);
   tp = (tcp_Header *)((byte *)ip + len);  /* TCP frame pointer */
   len = ntohs(ip->length) - len;  /* Len of TCP data */
   flags = ntohs(tp->flags);

   if (debug_on >= 1) {
      /* Display stuff on the far right of the second screen line */
      ++mono[310];  ++colour[310];
      mono[312] = colour[312] = (flags & tcp_FlagSYN) ? 'S' : ' ';
      mono[314] = colour[314] = (flags & tcp_FlagACK) ? 'A' : ' ';
      mono[316] = colour[316] = (flags & tcp_FlagFIN) ? 'F' : ' ';
      mono[318] = colour[318] = (flags & tcp_FlagRST) ? 'R' : ' ';
      }

   /* Demux to active sockets */
   for (s = tcp_allsocs; s; s = s->next) {
      if (s->safetysig != SAFETYTCP) {
         if (debug_on) outs("chain error in tcp\r\n");
         }
      if (s->state > tcp_StateLISTEN &&
            ntohs(tp->dstPort) == s->myport &&
            ntohs(tp->srcPort) == s->hisport &&
            ntohl(ip->destination) == s->myaddr &&
            ntohl(ip->source) == s->hisaddr)
         break;
      }

   if (!s && (flags & tcp_FlagSYN)) {
      /* Demux to passive sockets */
      for (s = tcp_allsocs; s; s = s->next) {
         if (s->state == tcp_StateLISTEN &&
               ntohs(tp->dstPort) == s->myport &&
               (s->hisport == 0 || ntohs(tp->srcPort) == s->hisport) &&
               (s->hisaddr == 0 || ntohl(ip->source) == s->hisaddr)) {
            s->myaddr = ntohl(ip->destination);
            break;
            }
         }
      }

   if (!s) {
      if (!(flags & tcp_FlagRST)) tcp_rst(ip, tp);
      return ok_to_free;
      }
   if (_dbugrecv) (*_dbugrecv)((sock_type *)s,ip,tp, "tcp_handler()");

   ph.src = ip->source;  /* Already INTELled */
   ph.dst = ip->destination;
   ph.mbz = 0;
   ph.protocol = IPPROTO_TCP;
   ph.length = htons(len);
   ph.checksum =  checksum(tp, len);
   if (checksum(&ph, sizeof(ph)) != 0xFFFF) {
#if defined(__FLAT__)
      if (debug_on)
         printf("Bad tcp checksum for packet at log/phys %.8X/%.8X len %X\n",
            tp, pkt_flat2phys(tp), len);
#endif
#ifdef N_DEBUG_SOON
      printf("sendsoon from tcp_handler()\n");
#endif
      tcp_sendsoon(s);
      return ok_to_free;
      }

   if (flags & tcp_FlagRST) {  /* Reset code */
      if (debug_on) outs("\7\7\7\7\7\7\7Connection reset\n");
      s->rdatalen = 0;
      s->rdataoffset = 0;
      s->err_msg = "Remote reset connection";
      s->state = tcp_StateCLOSED;
      tcp_abort(s);
      return ok_to_free;
      }

   if (sock_inactive)
        s->inactive_to = set_timeout(sock_inactive);

   /* Update our retransmission stuff */

   pkt_ack = ntohl(tp->acknum);
   pkt_seq = ntohl(tp->seqnum);
   if (s->tx_time > 0 && pkt_ack >= s->tx_seq) {
      M = set_ttimeout(0) - s->tx_time;  /* VJ's M (measured rtt) */
#ifdef N_DEBUG_RTT
      M_before = M;
#endif
      if (!s->AD_ok) {  /* First measurement */
         s->sA = M+9 << 3;  /* A = M + 0.5s, Stevens p305 */
         s->sD = s->sA >> 2;  /* D = A/2 */
         s->AD_ok = 1;
         }
      else {  /* VJ rtt estimate */
         M -= s->sA >> 3;
         s->sA += M;
         if (M < 0) M = -M;
         M -= s->sD >> 2;
         s->sD += M;  /* D = A + 4D */
         }
      s->rto = (s->sA >> 3) + s->sD;
#ifdef N_DEBUG_RTT
      printf("tcp_handler():  seq=%lu M=%d caused VJ(D_ok=%u sA=%u sD=%u rto=%u)\n",
         pkt_ack-s->dbg_tx_seqbase, M_before, s->AD_ok, s->sA, s->sD, s->rto);
#endif
      s->tx_time = 0;
      }
   else {  /* rtt timer not set; can't update A and D */
      }

#if defined(DEBUG)
   if (debug_on > 1)
      printf("-- rto=%u  sa=%u  sd=%u   cwnd=%u  wwnd=%u  wwnd2=%u  unacked=%lu\n",
         s->rto, s->sA, s->sD, s->cwnd, s->wwnd, s->wwnd2, s->unacked);
#endif

   if (flags & tcp_FlagACK) {
#if defined(N_DEBUG_RECV)
      if (pkt_ack != lastack) {
         lastack = pkt_ack;  lastack_count = 1;
         }
      else if (pkt_ack-s->dbg_tx_seqbase > 1000) {  /* Some data received */
         if (++lastack_count == 50) {
            s->err_msg = "50 dupl acks received";
            exit(7);
            }
         }
#endif
      pkt_window = ntohs( tp->window );
      if (!(flags & tcp_FlagSYN))
         pkt_window <<= s->rx_wscale;
      if (pkt_ack == s->seqnum &&        /* It's a duplicate ack */
            s->window == pkt_window &&   /* advertized window hasn't changed */
            s->sdatalen != 0) {          /* We have data waiting for an ack */
         ++s->dup_ack;
         }
      else {  /* Ack for new data */
         if (s->dup_ack > 3)
            /* Fast recovery, Stevens vol 1, p 312 and vol 2, p 974 */
            if (s->cwnd > s->ssthresh)
               s->cwnd = s->ssthresh;
         s->dup_ack = 0;
         }
      }
#ifdef N_DEBUG_CA
   if (s->dup_ack)
      printf("Duplicate ACK: ssthresh=%u, cwnd=%u, dup_ack=%u\n",
         s->ssthresh, s->cwnd, s->dup_ack);
#endif

   switch (s->state) {
   case tcp_StateLISTEN:   /* Accepting SYNs */
      /* Save his Ethernet address */
      if (_pktipofs)
         movmem(&((((eth_Header *)ip) - 1)->source[0]),
            &s->hisethaddr[0], sizeof(eth_address));
      if (flags & tcp_FlagSYN) {
         if (ip->tos > s->tos)
            s->tos = ip->tos;
         else if (ip->tos < s->tos) {
            /* RFC 793 says we should close connection */
            /* We best not do that while SunOS ignores TOS */
            }
         s->dbg_rx_seqbase = pkt_seq;
         s->acknum = pkt_seq+1;
         s->hisport = ntohs(tp->srcPort);
         s->hisaddr = ntohl(ip->source);
         s->flags = tcp_FlagSYN | tcp_FlagACK;
         s->state = tcp_StateSYNREC;
         tcp_ProcessOptions(s, tp, len);
         /* We must respond immediately */
         tcp_send(s, SEND_NORMAL, "tcp(LISTEN)");
         /* They must respond eventually */
         s->abort_timeout = set_timeout(tcp_OPEN_TIMEOUT);
         }
      return ok_to_free;

   case tcp_StateSYNSENT:
      if (flags & tcp_FlagSYN) {
         if (ip->tos > s->tos)
            s->tos = ip->tos;
         else if (ip->tos < s->tos) {
            /* RFC 793 says we should close connection */
            /* we best not do that while SunOS ignores TOS */
            }

         s->flags = tcp_FlagACK;
         s->abort_timeout = set_timeout(tcp_TIMEOUT);

         /* FlagACK means connection established, else SYNREC */
         if (flags & tcp_FlagACK) {
            /* But is it for the correct session ? */
            if (pkt_ack == s->seqnum+1) {
               s->state = tcp_StateESTAB;
               ++s->seqnum;    /* Good increment */
               s->acknum = pkt_seq+1;  /* 32 bits */
               s->dbg_rx_seqbase = pkt_seq;
               /* Someone may try it */
               tcp_ProcessData(s, &acked, &ok_to_free, ip, tp, len);
               if (!acked)
                  tcp_send(s, SEND_NORMAL, "tcp(SYNSENT)");
               }
            else {
               /* Wrong ack, force a RST and resend SYN soon*/
               s->flags = tcp_FlagRST;
               tcp_send(s, SEND_NORMAL, "tcp(SYN:RST)");
               s->flags = tcp_FlagSYN;
               tcp_send(s, SEND_NORMAL, "tcp(SYN:SYN)");
               }
            }
         else {
            s->acknum = pkt_seq+1;  /* 32 bits */
            s->dbg_rx_seqbase = pkt_seq;
            s->flags = tcp_FlagSYN | tcp_FlagACK;
            s->state = tcp_StateSYNREC;
            tcp_ProcessOptions(s, tp, len);
            tcp_send(s, SEND_NORMAL, "tcp(simultaneous open)");
              /* Must respond immediately */
            s->abort_timeout = set_timeout(tcp_TIMEOUT);
            }
         }
      return ok_to_free;

   case tcp_StateSYNREC:   /* rec SYN, sent SYN ACK, waiting ACK */
      if (flags & tcp_FlagSYN) {
         s->flags = tcp_FlagSYN | tcp_FlagACK;
         tcp_ProcessOptions(s, tp, len);
         tcp_send(s, SEND_NORMAL, "tcp(SYNREC)");
         s->abort_timeout = set_timeout(tcp_TIMEOUT);
         return ok_to_free;
         }
      if ((flags & tcp_FlagACK) && (pkt_ack == s->seqnum+1)) {
         s->flags = tcp_FlagACK;
         s->state = tcp_StateESTAB;
         ++s->seqnum;
         s->abort_timeout = 0;  /* Never timeout */
         tcp_ProcessOptions(s, tp, len);
         s->window = ntohs(tp->window) << s->rx_wscale;

         /* Having got to ESTAB, we must check for data in this packet.
          *   We do this by flowing into the next case (below) */
         }
      else return ok_to_free;

   case tcp_StateESTAB:
   case tcp_StateESTCL:
   case tcp_StateCLOSWT:
   case tcp_StateCLOSWTCL:

      /* Handle lost SYN ACK */
      if ((flags & tcp_FlagSYN) && (flags & tcp_FlagACK)) {
         /* don't process options because it is too late
          * to change MSS and WSCALE */
         tcp_send(s, SEND_NORMAL, "tcp(ESTAB:lost)");
         return ok_to_free;
         }
      /* Handle SYN without ACK */
      if (flags & tcp_FlagSYN) {
         tcp_rst(ip, tp);
         return ok_to_free;
         }
      /* Must have something to ack */
      if (!(flags & tcp_FlagACK))
         return ok_to_free;
      s->abort_timeout = 0;  /* We do not use this timeout at this point */
         /* Will be set again by ProcessData if we receive some data */

      /* Process ack value in packet .. */

      acked = 0;
      if ((r = seqcmp(pkt_ack,s->seqnum)) > 0) {
         diff = pkt_ack - s->seqnum;
         if (diff <= s->maxunacked) {  /* Ack of (sent but) unacked data */
            s->maxunacked -= diff;
            s->sdatalen -= diff;
            if (s->unacked > diff) s->unacked -= diff;
            else s->unacked = 0;
            s->sdataoffset = wrap(s->sdataoffset+diff, s->maxsdatalen);
            s->seqnum += diff;
                             /* SS/CA code follows:  Nevil, 30 Dec 97 */
            if (s->cwnd <= s->ssthresh)  /* In slow start */
               ++s->cwnd;
            else {  /* Doing congestion avoidance */
               if (++s->wwnd == s->cwnd) {
                  ++s->cwnd;
                  s->wwnd = 0;
                  }
               if (++s->wwnd2 == 8) {           /* Stevens p. 310 */
                  ++s->cwnd;
                  s->wwnd2 = 0;
                  }
               }
#ifdef N_DEBUG_CA
            printf("tcp_StateESTAB: ssthresh=%u, cwnd=%u, wwnd=%u, wwnd2=%u\n",
               s->ssthresh, s->cwnd, s->wwnd, s->wwnd2);
#endif
            }
         else {  /* Ack of unsent data */
            /* This is an error!  Just ignore the ack */
#ifdef DEBUG
            if (debug_on >1)
               printf("tcphandler confused, but didn't set unacked back to 0 from %lu\n",s->unacked);
#endif DEBUG
            }
         }
      else if (r == 0 && s->dup_ack != 0) {  /* Zero-length ack */
         if (s->dup_ack == 3) {  /* Fast recovery; Stevens v1 p312, v2 pp972-973 */
#ifdef J_DEBUG_DUP_ACK
            printf("3rd dup ack for %lu\n", pkt_ack - s->dbg_tx_seqbase);
#endif
            set_ssthresh(s);
            s->cwnd += 3;
            tcp_send(s, SEND_FAST_RETRANSMIT,
               "tcp(ESTAB: fast retransmit)");
            acked = 1;
            }
         else if (s->dup_ack > 3) {  /* First retrans didn't get through! */
            s->cwnd++;
            tcp_send(s, SEND_IF_HAVE_DATA, "tcp(ESTAB: fast retrans 2)");
            acked = 1;
            }
         }
      /* else ack < seq.  Already acked, ignore the ack */

      s->flags = tcp_FlagACK;
      rx_len = tcp_ProcessData(s, &acked, &ok_to_free, ip, tp, len);
      if (rx_len) {  /* We received some data */
         ++s->data_pkts;
         if (s->rdatalen >= s->maxrdatalen - 2*s->mss)
            *rdata_avail = 1;  /* Rx window getting close to full */
         }

      if ((flags & tcp_FlagFIN) &&
            (s->state != tcp_StateCLOSWT) &&
            (s->state != tcp_StateCLOSWTCL) &&
            (s->acknum == pkt_seq)) {
         ++s->acknum;  /* FIN counts as 1 */
         if (!s->err_msg) s->err_msg = "Connection closed";
         s->state = tcp_StateCLOSWT;
         tcp_send(s, SEND_NORMAL, "tcp(ESTAB:CLOS)");
         /* don't transition to LASTACK yet;
          * let user's call to tcp_close() cause that
          * Joel 5 June 1998
          */
         }

/* Send **NOW** when:
 *   tx: got (first) ack of unacked data && have > mss/2 available to send
 *       (sendable_data() allows for changes in his window size)
 *   rx: received new data && have room for more
 *                                 Nevil, 7 Jan 98 */

      if (!acked) {
         sent = 0;
         should_send = 0;
         if (r>0 && diff>0) {  /* Some sent data was acked */
            should_send = 1;
            if (sendable_data(s) > (longword)s->mss/2) {  /* We have data to send */
               tcp_send(s, SEND_NORMAL, "tcp(ESTAB: data to send");
               sent = 1;
               }
            }
         if (!sent && rx_len > 0) {  /* Some data was received */
            should_send = 1;
            if (s->data_pkts >= 2) {  /* ack 'now' for every 2nd packet */
               tcp_send(s, SEND_NORMAL, "tcp(ESTAB: space to receive");
               s->data_pkts = 0;
               sent = 1;
               }
            }
         if (!sent && should_send) {
#ifdef N_DEBUG_SOON
            printf("sendsoon from StateESTAB:\n");
#endif
            tcp_sendsoon(s);
            }
         }

      if (s->state == tcp_StateESTCL ||
            s->state == tcp_StateCLOSWTCL)
         tcp_close(s);
      return ok_to_free;

   case tcp_StateFINWT1:
      /* They have not necessarily read all the data yet,
            we must still supply it as requested */

      /* We waited for them to acknowledge all our data (which causes
       * us to delete same from our transmit buffer), using the ESTCL
       * state, before sending FIN, so doesn't it indicate an error
       * on their part to send any acknum other than exactly same as
       * our sequence number???
       *
       * We get away with not checking result of seqcmp to be
       * less than zero, to ensure I don't get a negative diff
       * when their ack is beyond our sequence number, because
       * a negative diff would look like large positive, which
       * would fail to be less than s->maxunacked.
       *
       * Joel   5 June 1998
       */

#ifdef N_DEBUG_UPDOWN
      printf("rcv FINWT1: tp->acknum=%lu, diff=%d, s->sdatalen=%d, len=%d, flags=%x\n",
         pkt_ack, diff, s->sdatalen, len, flags);
#endif
      if (seqcmp(pkt_ack,s->seqnum)) {
         diff = pkt_ack - s->seqnum;
         if (diff <= s->maxunacked) {
            s->maxunacked -= diff;
            s->sdatalen -= diff;
            if (s->unacked > diff) s->unacked -= diff;
            else s->unacked = 0;
            s->sdataoffset = wrap(s->sdataoffset + diff, s->maxsdatalen);
            s->seqnum += diff;
            }
         }

      /* They may still be transmitting data, we must read it */
      tcp_ProcessData(s, &acked, &ok_to_free, ip, tp, len);

      /* Check if other tcp has acked all sent data and is
            ready to change states */

      if (flags & tcp_FlagFIN) {
         /* Trying to do simultaneous close */
         if (flags & tcp_FlagACK) {
#ifdef N_DEBUG_UPDOWN
            printf("FINWT1: simultaneous close with ACK\n");
#endif
            if ((pkt_ack >= s->seqnum + 1)
                   && (pkt_seq == s->acknum)) {
               ++s->seqnum;  /* He's acked our fin */
               ++s->acknum;
               s->flags = tcp_FlagACK;
               tcp_send(s, SEND_NORMAL, "tcp(FINWT1)");
               s->abort_timeout = set_timeout(tcp_TIMEWT_TIMEOUT);
               s->state = tcp_StateTIMEWT;
               }
            }
         else {
#ifdef N_DEBUG_UPDOWN
            printf("FINWT1: simultaneous close without ACK\n");
#endif
            if (pkt_seq == s->acknum) {
               ++s->acknum;
               s->flags = tcp_FlagACK;
               tcp_send(s, SEND_NORMAL, "tcp(FINWT1)");
               s->abort_timeout = tcp_CLOSING_TIMEOUT ? set_timeout(tcp_CLOSING_TIMEOUT) : 0;
               s->state = tcp_StateCLOSING;
               }
            }
         }
      else if (flags & tcp_FlagACK) {
#ifdef N_DEBUG_UPDOWN
         printf("FINWT1: acking our fin\n");
#endif
         /* Other side is legitimately acking our fin */
         if ((pkt_ack == s->seqnum + 1) &&
               (pkt_seq == s->acknum) &&
               (s->sdatalen == 0)) {
            ++s->seqnum;
            /* They are just acking our seq num, not sending more data
             *    more data for us to ack */
            /*       s->acknum++;  Do it in FINWT2.  Nevil, 22 Apr 93 */
            s->state = tcp_StateFINWT2;
            s->abort_timeout = tcp_FINWT2_TIMEOUT ? set_timeout(tcp_FINWT2_TIMEOUT) : 0;
            }
         }
      break;

   case tcp_StateFINWT2:
#ifdef N_DEBUG_UPDOWN
      printf("rcv FINWT2: s->sdatalen=%d, len=%d, flags=%x\n", s->sdatalen, len, flags);
#endif
      /* They may still be transmitting data, we must read it */
      tcp_ProcessData(s, &acked, &ok_to_free, ip, tp, len);

      if ((flags & (tcp_FlagACK | tcp_FlagFIN)) ==
            (tcp_FlagACK | tcp_FlagFIN)) {
         if ((pkt_ack == s->seqnum) &&
               (pkt_seq == s->acknum)) {
            ++s->acknum;  /* Now we have his fin */
            s->flags = tcp_FlagACK;
            tcp_send(s, SEND_NORMAL, "tcp(FINWT2)");
            s->abort_timeout = set_timeout(tcp_TIMEWT_TIMEOUT);
            s->state = tcp_StateTIMEWT;
            return ok_to_free;
            }
         }
      break;

   case tcp_StateCLOSING:
#ifdef N_DEBUG_UPDOWN
      printf("rcv CLOSING: s->sdatalen=%d, len=%d, flags=%x\n", s->sdatalen, len, flags);
#endif
      if ((flags & (tcp_FlagACK | tcp_FlagFIN)) == tcp_FlagACK) {
         if ((pkt_ack >= s->seqnum + 1)
               && (pkt_seq == s->acknum)) {
            s->state = tcp_StateTIMEWT;
            s->abort_timeout = set_timeout(tcp_TIMEWT_TIMEOUT);
            }
         }
      break;

   case tcp_StateLASTACK:
#ifdef N_DEBUG_UPDOWN
      printf("rcv LASTACK: s->sdatalen=%d, len=%d, flags=%x\n",
         s->sdatalen,len, flags);
#endif
      if (flags & tcp_FlagFIN) {
         /* They lost our two packets, back up */
         s->flags = tcp_FlagACK | tcp_FlagFIN;
         tcp_send(s, SEND_NORMAL, "tcp(LASTACK)");
         return ok_to_free;
         }
      else {
         if ((pkt_ack == s->seqnum+1) &&
               (pkt_seq == s->acknum)) {
            s->state = tcp_StateCLOSED;  /* No 2msl necessary */
            s->abort_timeout = 0;
            return ok_to_free;
            }
         }
      break;

   case tcp_StateTIMEWT:
#ifdef N_DEBUG_UPDOWN
      printf("rcv TIMEWT: s->sdatalen=%d, len=%d, flags=%x\n",
      s->sdatalen,len, flags);
#endif
      if ((flags & (tcp_FlagACK | tcp_FlagFIN)) == (tcp_FlagACK | tcp_FlagFIN)) {
         /* he needs an ack */
         s->flags = tcp_FlagACK;
         tcp_send(s, SEND_NORMAL, "tcp(TIMEWT)");
         /* but we still don't want to change state */
         }
      break;
      }

   return ok_to_free;
   }

static word move_data_to_rx_buffer(tcp_Socket *s, tcp_Header *tp,
   byte *dp, word len, Bit32 offset, int is_in_order)
{
   longword x, startpos, up_to_end;

   if (is_in_order && s->rdataHandler) {
      len = s->rdataHandler(s, dp, len, 0, tp);
      s->acknum += len;
      }
   else {
      /* No handler or out of order, just dump to buffer,
       * should be indexed, handles goofs */
      /* Limit receive size to our window */

      if (len > (x = s->maxrdatalen - s->rdatalen - offset))
         len = x;
      if (len > 0) {
         startpos = wrap(s->rdatalen + s->rdataoffset + offset, s->maxrdatalen);
         up_to_end = s->maxrdatalen - startpos;
         if (up_to_end < len) {
            movmem(dp, s->rdata + startpos, up_to_end);
            movmem(dp+up_to_end, s->rdata, len-up_to_end);
            }
         else movmem(dp, s->rdata + startpos, len);
         if (is_in_order) {
            s->acknum += len;  /* Our new ack begins at end of data */
            s->rdatalen += len;
            }
         /* Note: data offset doesn't change when we add to buffer */
         }
      }
   return len;
   }

/* assumes offset is zero
 *
 * passes NULL to datahandler as TCP header pointer,
 * which might be a problem???  (hey, I don't have a
 * real packet to pass, and I don't know if caller
 * expects to find data after that header, so this
 * has the best chance of flushing out their intentions!!!)
 */
static longword fake_data_to_rx_buffer(tcp_Socket *s, Bit32 len)
{
   longword x, startpos, up_to_end;
   byte *dp;

   if (s->rdataHandler) {
      startpos = wrap(s->rdatalen + s->rdataoffset, s->maxrdatalen);
      up_to_end = s->maxrdatalen - startpos;
      if (up_to_end < len) {
         x = s->rdataHandler(s, s->rdata+startpos, up_to_end, 0, NULL);
         if (x == up_to_end)
            x += s->rdataHandler(s, s->rdata, len-up_to_end, 0, NULL);
         len = x;
         }
      else
         len = s->rdataHandler(s, s->rdata+startpos, len, 0, NULL);
      }
   else
      s->rdatalen += len;
   s->acknum += len;  /* Our new ack begins at end of data */
   return len;
   }

/* length includes TCP header, TCP options, and TCP payload */
static word tcp_ProcessOptions(tcp_Socket *s, tcp_Header *tp, word len)
{
   word x;
   int numoptions;          // can go negative when bad option
   word opt_temp;
   word opt_len;
   byte *options;
   word is_syn = ntohs(tp->flags) & tcp_FlagSYN;

   /* Find the data portion */
   x = tcp_GetDataOffset(tp) << 2;  /* convert quads to bytes */
   /* TCP header or TCP options could have been truncated by IP */
   if (x > len)
      x = len;
   /* Process those options */
   if ((numoptions = x - sizeof(tcp_Header)) > 0) {
      options = (byte *)tp + sizeof(tcp_Header);
      while (numoptions-- > 0) {
         switch (*options++) {
         case  0:  /* End of options */
            numoptions = 0;
            break;
         case  1 :  /* Nop */
            break;
         case  2:   /* MSS */
            opt_len = *options;
            if (opt_len==4 && is_syn) {
               opt_temp = ntohs(*(word*)(&options[1]));
               if (opt_temp < s->mss)
                  s->mss = opt_temp;
               }
            numoptions -= opt_len - 1;
            options += opt_len - 1;
            break;
         case  3:   /* window scaling */
            opt_len = *options;
            if (opt_len==3 && is_syn && s->rx_wscale==BIT8_MAX) {
               s->rx_wscale = options[1];
               if (s->rx_wscale > tcp_MaxWinShift)
                  s->rx_wscale = tcp_MaxWinShift;
               }
            numoptions -= opt_len - 1;
            options += opt_len - 1;
            break;
            }
         }
      }

   /* if the connection will be established after this packet
    * and the window scale option hasn't been received yet,
    * it is time to force both TX and RX windows to be unscaled
    */
   if (s->state==tcp_StateESTAB && s->rx_wscale==BIT8_MAX) {
      s->rx_wscale = 0;
      s->tx_wscale = 0;
      }

   return x;
   }

/*
 * Process the data in an incoming packet.
 * Called from all states where incoming data can be received: estab,
 * fin-wait-1, fin-wait-2
 *
 * Returns the amount of new TCP data that we accepted from this packet
 */
static word tcp_ProcessData(tcp_Socket *s, int *acked, int *ok_to_free,
   in_Header *ip, tcp_Header *tp, word len)
{
   Bit32 diff;
   Bit32 pkt_seq;
   Bit32 biglen;
   int r, ok;
   word x;
   word flags;
   byte *dp;
   struct pseg *bp;
#ifdef J_DEBUG_RX
   extern long nobufpackets;
#endif

   x = tcp_ProcessOptions(s, tp, len);
   len -= x;              /* Remove the header length */
   dp = (byte *)tp + x;

#if defined(J_DEBUG_RX)
   printf("<- ProcessData: p->seq=%lu, len=%u, p->ack=%lu, p->wnd=%u, ready=%ld\n",
      ntohl(tp->seqnum)-s->dbg_rx_seqbase, len, ntohl(tp->acknum)-s->dbg_tx_seqbase, ntohs(tp->window), readyq.len );
#endif

   pkt_seq = ntohl( tp->seqnum );
   flags = ntohs(tp->flags);
   if ((flags & tcp_FlagSYN) == 0)
      s->window = ntohs(tp->window) << s->rx_wscale;
   else
      s->window = ntohs(tp->window);


   /* Done option processing */
   if ((r = seqcmp(pkt_seq,s->acknum)) == 0)
      ok = 1;
   else {
      ok = 0;
      if (r < 0) {  /* seq < ack */
         diff = s->acknum - pkt_seq;
         if (flags & tcp_FlagSYN) {
            /* SYN uses 1 byte of sequence space (Stevens, p231) */
            --diff;
            ok = 1;
            }
         if (diff < len) {  /* Acked byte not first in packet */
            dp += diff;  /* Skip bytes already received */
            len -= diff;
            ok = len > 0;
            }
         }
      }
   if (ok) {  /* Correct sequence */
      len = move_data_to_rx_buffer(s,tp, dp,len,0, TRUE);
      if (s->misordered.len != 0) {  /* Out-of-sequence pkts queued */
         /* if first segment starts before or at current packet end... */
         bp = seg_dequeue_head( &s->misordered );
         if ((r = seqcmp(bp->seq,s->acknum)) <= 0) {
            biglen = bp->len;
            diff = s->acknum - bp->seq;
#ifdef N_DEBUG_RECV
            printf("+++ seg %u dqueued: seq=%lu, ack=%lu, len=%lu, diff=%lu\n",
               s->misordered.len+1,
               bp->seq-s->dbg_rx_seqbase, s->acknum-s->dbg_rx_seqbase,
               biglen, diff);
#endif
            if (diff <= biglen) {  /* Some good new data in queued pkt */
               biglen -= diff;
               }
            else biglen = 0;  /* Incoming packet overlaps queued seg completely */
            if (biglen > 0)
               fake_data_to_rx_buffer( s, biglen );
            seg_enqueue_head( &misordered_free, bp );
            }
         else {
            // put it back in front of misordered queue
            //
            seg_enqueue_head( &s->misordered, bp );
            }
         }
      }
   else {  /* Not ok */
      if (r > 0) {  /* We have a gap in the packet sequence */
         tcp_send(s, SEND_IMMEDIATE_ACK, "ProcData(Missing)");
         *acked = 1;
#ifndef J_TEST_MISORDERING
         bp = seg_dequeue_head( &misordered_free );
         if (bp) {
            diff = pkt_seq - s->acknum;
            len = move_data_to_rx_buffer(s,tp, dp,len,diff, FALSE);
            bp->seq = pkt_seq;
            bp->len = len;
            seg_enqueue_merge( &s->misordered, bp );

#ifdef N_DEBUG_RECV
            printf("--- pkt %u queued: seq=%lu, ack=%lu, len=%lu, diff=%lu\n",
               s->misordered.len, bp->seq-s->dbg_rx_seqbase, s->acknum-s->dbg_rx_seqbase, bp->len, diff);
#endif
            }
         else {
            fprintf( stderr, "tcp_ProcessData() misordered segment free"
               " list(%u) empty!\n", MAX_MISORDERED );
            }
#endif // !defined( J_TEST_MISORDERING )
         *ok_to_free = 1;
         }
      else {  /* r < 0:  repeated seqnbr */
         if (len == 0) {  /* Ask for retransmit */
            tcp_send(s, SEND_IMMEDIATE_ACK, "ProcData(Repeated)");
            *acked = 1;
            }
         /* else repeated data; ignore it */
         }
      }

   // If we just sucked up some data
   // and this is a FIN packet...
   //              Joel Apisdorf 7/2/96
   //
   if ((len > 0) && (flags & tcp_FlagFIN)) {
      // Fudge the sequence number in the packet
      // so our caller can assume no data was attached
      // when he decides whether he has received it all,
      // since we are the only code that really knows
      // how much was processed this time around
      //
      // This would not be necessary if we were willing
      // to send out an extra acknowledge of just-received
      // data (possibly with FIN bit also set), causing next
      // transmission (probably with FIN bit set) from other
      // side to have a sequence number our caller would like,
      // but that would cost an extra round-trip time
      //
      tp->seqnum = htonl(len + ntohl(tp->seqnum));
      }

   s->abort_timeout = set_timeout(tcp_TIMEOUT);
   return len;
   }

/*
 * Format and send an outgoing segment
 */
static void tcp_send(tcp_Socket *s, int type, char *from)
{
   tcp_PseudoHeader ph;
   struct _pkt {
      in_Header in;
      tcp_Header tcp;
      word maxsegopt[2];
      byte nopopt[1];
      byte wscaleopt[3];
      } *pkt;
   struct _tunnel {
      in_Header  in;
      struct _pkt pkt;
      } *tunnel;
   byte *dp;
   in_Header *inp;
   tcp_Header *tcpp;
   word senddatalen, sendpktlen;
   longword startdata, sendtotdata;
#if defined(DEBUG)
   longword sendtotlen;
#endif

   int ippkt, mxpkt;
   longword up_to_end, abs_startpos;

   if (s->tunneladdr) {
      tunnel = (struct _tunnel *)_eth_formatpacket(
         &s->hisethaddr[0], /*0x0800*/ 8);
      pkt = &tunnel->pkt;
      }
   else {
      pkt = (struct _pkt *)_eth_formatpacket(&s->hisethaddr[0], /*0x800*/ 8);
      }
   dp = (byte *) &pkt->maxsegopt;  /* dp constant for multi-packet sends */
   inp = &pkt->in;
   tcpp = &pkt->tcp;

   if (type==SEND_NORMAL || type==SEND_IMMEDIATE_ACK || type==SEND_IF_HAVE_DATA) {
      startdata = s->unacked;
      sendtotdata = sendable_data(s);  /* Total possible send size */
      if (type==SEND_IF_HAVE_DATA && sendtotdata==0)
         return;
      }
   else if (type == SEND_FAST_RETRANSMIT) {
      startdata = 0;    /* Don't check his window here .. we're */
      sendtotdata = s->sdatalen;  /*   re-sending data inside it! */
      }

#if defined(DEBUG)
   sendtotlen = 0;     /* Running count of what we've sent */
#endif
#ifdef N_DEBUG_SEND
   printf("send(%s): datalen=%lu, unacked=%lu, window=%lu -> tot=%lu\n",
      from, s->sdatalen, s->unacked, s->window, sendtotdata);
#endif

   /* Step through our packets */
   for (ippkt = 0; ;) {
      /* adjust size for each packet
       *
       * always results in a value less than 64K
       * even though sendtotdata is 32 bits,
       * because s->mss is only 16 bits
       *
       * Joel 13 May 1998
       */
      if (s->tunneladdr)
         senddatalen = min(sendtotdata, s->mss-sizeof(in_Header));
      else
         senddatalen = min(sendtotdata, s->mss);

      /* TCP header */
      tcpp->srcPort = htons(s->myport);
      tcpp->dstPort = htons(s->hisport);
      tcpp->seqnum = htonl(s->seqnum+startdata);
      tcpp->acknum = htonl(s->acknum);
      s->oldwindow = calc_rx_window(s);
      tcpp->window = htons(s->oldwindow);
#if defined(J_DEBUG_TX)
      printf("<- send:        p->ack=%lu, p->seq=%lu, p->window=%u\n",
         ntohl(tcpp->acknum), ntohl(tcpp->seqnum), ntohs(tcpp->window));
#endif
      tcpp->flags = htons(s->flags | 0x5000);
      tcpp->checksum = 0;
      tcpp->urgentPointer = 0;

      /* Do options if this is our first packet */
      if (s->flags & tcp_FlagSYN) {
         pkt->maxsegopt[0] = htons(0x0204);
         pkt->maxsegopt[1] = htons(s->mss);
         /* only send window scaling option if this is active-open,
          * or if this is passive-open that received wscale option */
         if (s->state==tcp_StateSYNSENT ||
            (s->state==tcp_StateSYNREC && s->rx_wscale!=BIT8_MAX)) {
            sendpktlen = sizeof(struct _pkt);
            tcpp->flags = htons(ntohs(tcpp->flags) + 0x2000);
            dp += 8;
            pkt->nopopt[0] = 1;                    // no operation
            pkt->wscaleopt[0] = 3;                 // option ID
            pkt->wscaleopt[1] = 3;                 // byte length incl. all bytes
            pkt->wscaleopt[2] = s->tx_wscale;      // window shift count
            }
         else {
            sendpktlen = sizeof(struct _pkt) - 4;
            tcpp->flags = htons(ntohs(tcpp->flags) + 0x1000);
            dp += 4;
            }
         }
      else {  /* Handle packets with data */
         if (senddatalen > 0) {
            sendpktlen = senddatalen + sizeof(tcp_Header) + sizeof(in_Header);

            /* Get data from appropriate place */
#ifdef J_DEBUG_TCP_SEND_STARTDATA
            printf("tcp_send() ippkt=%d startdata=%lu sendatalen=%u\n", ippkt, startdata, senddatalen);
#endif
            abs_startpos = wrap(s->sdataoffset+startdata, s->maxsdatalen);
            up_to_end = s->maxsdatalen - abs_startpos;
            if (up_to_end < senddatalen) {
               movmem(s->sdata + abs_startpos, dp, up_to_end);
               movmem(s->sdata, dp+up_to_end, senddatalen-up_to_end);
               }
            else
               movmem(s->sdata + abs_startpos, dp, senddatalen);

            // Note: data offset doesn't change because we are not
            // removing data from buffer, because we might need to
            // retransmit it later
            //

/*          dp[ senddatalen ] = 0; */
            }
         else {
            /* Handle no-data, not-first-SYN packets */
                sendpktlen = sizeof(tcp_Header) + sizeof(in_Header);
            }
         }

      /* IP header */
      memset(inp, 0, sizeof(in_Header));
      inp->ver = 4;
      inp->hdrlen = 5;
      inp->tos = s->tos;
      inp->identification = htons(++ip_id);
//      inp->frag = 0;
      inp->ttl = s->ttl;
      inp->proto = IPPROTO_TCP;
      inp->checksum = 0;
      inp->source = htonl(s->myaddr);
      inp->destination = htonl(s->hisaddr);
      inp->length = htons(sendpktlen);

      inp->checksum = ~checksum(inp, sizeof(in_Header));

      /* Compute tcp checksum */
      ph.src = inp->source;   /* Already INTELled */
      ph.dst = inp->destination;
      ph.mbz = 0;
      ph.protocol = IPPROTO_TCP;
      ph.length = htons(sendpktlen - sizeof(in_Header));
      ph.checksum = checksum(&pkt->tcp, sendpktlen-sizeof(in_Header));

      tcpp->checksum = ~checksum(&ph, sizeof(ph));

      if (_dbugxmit) (*_dbugxmit)((sock_type *)s,inp,tcpp, "tcp_send()");
      if (debug_on >= 1) {
         /* Display stuff on the far right of the first screen line */
         mono[150]++;  colour[150]++;
         mono[152] = colour[152] = (s->flags & tcp_FlagSYN) ? 'S' : ' ';
         mono[154] = colour[154] = (s->flags & tcp_FlagACK) ? 'A' : ' ';
         mono[156] = colour[156] = (s->flags & tcp_FlagFIN) ? 'F' : ' ';
         mono[158] = colour[158] = (s->flags & tcp_FlagRST) ? 'R' : ' ';
         }

      if (s->tunneladdr) {
         inp = &tunnel->in;

         inp->ver = 4;
         inp->hdrlen = 5;
         inp->tos = 0;
         inp->identification = htons(++ip_id);
         inp->frags = 0;
         inp->fo = 0;
         inp->ttl = in_DefaultTTL;
         inp->proto = IPPROTO_IP_IN_IP;  /* Please unwrap me! */
         inp->source = htonl(s->myaddr);
         inp->destination = htonl(s->tunneladdr);
         inp->length = htons(sizeof(in_Header) + sendpktlen);
         inp->checksum = 0;
         inp->checksum = ~checksum(inp, sizeof(in_Header));
         }

      ++ippkt;
#ifdef N_TEST_TX_LOSS
      if (--tx_loss_int_ct == 0) {
         tx_loss_int_ct = tx_loss_interval + (hwt_last_count >> 8) % 17;
         if (tx_packet_count != 0) {
            tx_loss_nbr_ct =
               loss_array[(hwt_last_count >> 4) % sizeof(loss_array)];
            }
         }
      ++tx_packet_count;
      if (tx_loss_nbr_ct) {
         --tx_loss_nbr_ct;  /* Discard this packet */
         printf("****** TEST: ****** TX packet %lu discarded ******\n",
            tx_packet_count);
         }
      else  /* Send this one as normal */
#endif

      if (_eth_send(ntohs(inp->length))) {  /* Packet send failed */
         set_ssthresh(s);  /* Slow down transmitter */
#ifdef N_DEBUG_SOON
         printf("sendsoon from tcp_write()\n");
#endif
         tcp_sendsoon(s);  /* Try again soon */
         return;
         }

      /* Packet sent OK */
      if (type == SEND_FAST_RETRANSMIT || type == SEND_IMMEDIATE_ACK)
         break;

      startdata += senddatalen;
      if (s->tx_time == 0) {  /* rtt timer is available */
         s->tx_time = set_ttimeout(0);  /* Time last 'timed' packet sent */
         s->tx_seq = s->seqnum+startdata;
#ifdef N_DEBUG_RTT
         printf("timing packet with seq=%lu\n", s->tx_seq-s->dbg_tx_seqbase);
#endif
         }

      /* Update these here - that way it will be correct
         if send fails the next time around this loop */

      s->unacked = startdata;
      if (s->unacked > s->maxunacked)  /* Non-retransmitted data sent */
      {
#ifdef J_DEBUG_WINDOW_OVERFLOW
          printf( "tcp_send(%p,%d,%s) maxunacked %lu\n", s, type, from, s->unacked );
#endif
          s->maxunacked = s->unacked;
      }

      /* Do next ip pkt */
#if defined(DEBUG)
      sendtotlen += senddatalen;
#endif
      sendtotdata -= senddatalen;
      if (sendtotdata <= 0  ||  /* All available data sent */
           ippkt >= MAXPKTSEND)  /* Packet driver limit reached */
         break;
      }

#if defined(DEBUG)
   if (debug_on)
      printf(" Sent %lu/%lu bytes in %u/%u packets  "
         "with (%u) unacked SYN %lu  from %s\n",
         sendtotlen,s->window, ippkt,s->cwnd,
         s->unacked, s->seqnum, from);
#endif DEBUG

   s->retrans_timeout = set_ttimeout(s->rto);  /* Timeout from last pkt sent */
   s->soon_req = 0;
   }

/*
 * Format and send a reset tcp packet
 */
static void tcp_rst(in_Header *his_ip, tcp_Header *oldtcpp)
{
   tcp_PseudoHeader ph;
   struct _pkt {
      in_Header in;
      tcp_Header tcp;
      word maxsegopt[2];
      } *pkt;

   static longword nextrst = 0L;
   word oldflags;
   in_Header *inp;
   tcp_Header *tcpp;
   eth_Header *eth;
   word sendtotlen;  /* Length of packet */
   int temp;
   longword templong;

   /* See RFC 793 page 65 for details */

   if (!chk_timeout(nextrst)) return;
   nextrst = set_ttimeout(1);

   oldflags = ntohs(oldtcpp->flags);
   if (oldflags & tcp_FlagRST) return;
   if (oldflags & tcp_FlagACK) {
      oldtcpp->seqnum = oldtcpp->acknum;
      oldtcpp->acknum = 0;
      oldflags = tcp_FlagRST;
      }
   else if ((oldflags & (tcp_FlagSYN | tcp_FlagACK)) ==  tcp_FlagSYN) {
      oldtcpp->acknum = htonl(ntohl(oldtcpp->seqnum) + 1);
      oldtcpp->seqnum = 0;
      oldflags = tcp_FlagACK | tcp_FlagRST;
      }
   else {
      temp = ntohs(his_ip->length) - in_GetHdrlenBytes(his_ip);
      oldtcpp->acknum = htonl(ntohl(oldtcpp->seqnum) + temp);
      oldtcpp->seqnum = 0;
      oldflags = tcp_FlagRST;
      }

   /* Convoluted mechanism - reads his ethernet address or garbage */
   eth = _eth_hardware((byte *)his_ip);

   pkt = (struct _pkt *)_eth_formatpacket(eth, 8);
   inp = &pkt->in;
   tcpp = &pkt->tcp;

   sendtotlen = sizeof(tcp_Header) + sizeof(in_Header);
   memset(inp, 0, sizeof(in_Header));
   inp->length = htons(sendtotlen);

   /* TCP header */
   tcpp->srcPort = oldtcpp->dstPort;
   tcpp->dstPort = oldtcpp->srcPort;
   tcpp->seqnum = oldtcpp->seqnum;
   tcpp->acknum = oldtcpp->acknum;
   tcpp->window = 0;
      /*    tcpp->flags = htons(oldflags); */
      /* BUG FIX : jason dent found this thanks to SCO */
   tcpp->flags = htons((oldflags & 0x0fff) | 0x5000);
   tcpp->checksum = 0;
   tcpp->urgentPointer = 0;

   /* IP header */
   inp->ver = 4;
   inp->hdrlen = 5;
   inp->tos = his_ip->tos;
   inp->identification = htons(++ip_id);
//    inp->frag = 0;
   inp->ttl = in_DefaultTTL;
   inp->proto = IPPROTO_TCP;
   inp->checksum = 0;
   inp->source = his_ip->destination;
   inp->destination = his_ip->source;

   inp->checksum = ~checksum(inp, sizeof(in_Header))/* 0*/;

   /* Compute tcp checksum */
   ph.src = inp->source;  /* Already INTELled */
   ph.dst = inp->destination;
   ph.mbz = 0;
   ph.protocol = 6;
   ph.length = htons(sendtotlen - sizeof(in_Header));

   ph.checksum = checksum(&pkt->tcp, sizeof(tcp_Header));
   tcpp->checksum =  ~checksum(&ph, sizeof(ph));

   if (_dbugxmit) (*_dbugxmit)(NULL,inp,tcpp, "tcp_rst()");

   _eth_send(ntohs(inp->length));
   }

/**********************************************************************
 * Socket functions
 **********************************************************************/

/* Socket-based stuff */

/*
 * sock_read - Read a socket with maximum n bytes
 *           - Busywaits until buffer is full but calls s->usr_yield
 *           - Returns count also when connection gets closed
 */
longword sock_read(sock_type *s, byte *dp, longword len)
{
   char keystroke;
   int proto_is_tcp;
   longword templen, count;

   proto_is_tcp = (s->udp.ip_type == IPPROTO_TCP);
   if (!proto_is_tcp && s->udp.ip_type != IPPROTO_UDP)
      return 0;  /* Do nothing when socket is closed */

   count = 0;
   do {
      if (proto_is_tcp)
         templen = tcp_read(&s->tcp, dp, len);
      else
         templen = udp_read(&s->udp, dp, len);
      if (s->tcp.usr_yield)
         (s->tcp.usr_yield)();
      if (templen < 1) {
         if (!tcp_tick(s))
            break;
         }
      else {
         count += templen;
         dp += templen;
         len -= templen;
         }
      } while (len);
   return count;
   }

/*
 * Read a socket with maximum n bytes
 *
 * Does not busywait until buffer is full
 */
longword sock_fastread(sock_type *s, byte *dp, longword len)
{
   if (s->udp.ip_type == IPPROTO_UDP)
      return udp_read(&s->udp, dp, len);

   if (s->udp.ip_type == IPPROTO_TCP)
      return tcp_read(&s->tcp, dp, len);

   // Do nothing when socket is closed
   return 0;
   }

/*
 * sock_write - Writes data and returns length written
 *            - Does not perform flush
 *            - Repeatedly calls s->usr_yield
 */
longword sock_write(sock_type *s, byte *dp, longword len)
{
   longword templen, count;
   word oldmode;
   int proto_is_tcp;

   count = 0;

   proto_is_tcp = (s->udp.ip_type == IPPROTO_TCP);
   if (proto_is_tcp)
      oldmode = s->tcp.flags & tcp_FlagPUSH;
   else if (s->udp.ip_type != IPPROTO_UDP)
      return 0;  /* Do nothing when socket is closed */
   while (len  > 0) {
      if (proto_is_tcp) {
         s->tcp.flags |= oldmode;
         templen = tcp_write(&s->tcp, dp, len);
         }
      else
         templen = udp_write(&s->udp, dp, len);
      count += templen;
      dp += templen;
      len -= templen;
      if (s->udp.usr_yield)
         (s->udp.usr_yield)();
      if (!tcp_tick(s))
         break;
      }
   return count;
   }

longword sock_fastwrite(sock_type *s, byte *dp, longword len)
{
   if (s->udp.ip_type == IPPROTO_UDP)
      return udp_write(&s->udp, dp, len);

   if (s->udp.ip_type == IPPROTO_TCP)
      return tcp_write(&s->tcp, dp, len);

   // Do nothing when socket is closed
   return 0;
   }

/* Change whether default or user-allocated
 * buffer will be used for received data
 *
 * Safe to call after data has been received
 *
 * Will affect window scaling for TCP TX
 * if the SYN segment hasn't gone out yet
 */
longword sock_set_rx_buf(tcp_Socket *s, byte *dp, longword len)
{
   longword up_to_end;

   /* Socket must really be open */
   if (s->ip_type != IPPROTO_UDP
         && s->ip_type != IPPROTO_TCP)
      return 0;

   // Fix pointer and length if they want default buffering
   if (len == 0 || dp == NULL) {
      dp = s->rddata;

      if (s->ip_type == IPPROTO_TCP)
         len = tcp_MaxBufSize;
      else
         len = udp_MaxBufSize;
      }

   // If new buffer isn't big enough to hold existing
   // data, we must stop now, since it has already
   // been acknowledged
   if (s->rdatalen > len)
      return 0;

   // If there was any old data and pointer changed,
   // copy it to the new spot, at offset zero
   if (s->rdatalen && dp!=s->rdata) {
      up_to_end = s->maxrdatalen - s->rdataoffset;
      if (up_to_end < s->rdatalen) {
         memcpy(dp, s->rdata + s->rdataoffset, up_to_end);
         memcpy(dp+up_to_end, s->rdata, s->rdatalen - up_to_end);
         }
      else
         memcpy(dp, s->rdata + s->rdataoffset, s->rdatalen);

      s->rdataoffset = 0;
      }

   // Now we can change the pointer and length
   s->maxrdatalen = len;
   s->rdata = dp;

   // Now calculate the TCP RX window shift count using log2 of buffer length
   if (s->ip_type==IPPROTO_TCP && (s->state==tcp_StateOPEN || s->state==tcp_StateLISTEN)) {
      if( len < 65536 )
         s->tx_wscale = 0;
      else if( len < 131072 )
         s->tx_wscale = 1;
      else if( len < 262144 )
         s->tx_wscale = 2;
      else if( len < 524288 )
         s->tx_wscale = 3;
      else if( len < 1048576 )
         s->tx_wscale = 4;
      else if( len < 2097152 )
         s->tx_wscale = 5;
      else if( len < 4194304 )
         s->tx_wscale = 6;
      else if( len < 8388608 )
         s->tx_wscale = 7;
      else if( len < 16777216 )
         s->tx_wscale = 8;
      else if( len < 33554432 )
         s->tx_wscale = 9;
      else if( len < 67108864 )
         s->tx_wscale = 10;
      else if( len < 134217728 )
         s->tx_wscale = 11;
      else if( len < 268435456 )
         s->tx_wscale = 12;
      else if( len < 536870912 )
         s->tx_wscale = 13;
      else
         s->tx_wscale = 14;
      }

   // Successful!
   return s->maxrdatalen;
   }

/* Change whether default or user-allocated
 * buffer will be used for transmitted data
 *
 * Safe to call while data is queued
 *
 * Only works for TCP sockets, since only they
 * have buffers for output data
 */
longword sock_set_tx_buf(tcp_Socket *s, byte *dp, longword len)
{
   longword up_to_end;

   // Socket must really be open and must be TCP
   if (s->ip_type != IPPROTO_TCP)
      return 0;

   // Fix pointer and length if they want default buffering
   if (len == 0 || dp == NULL) {
      dp = s->sddata;
      len = tcp_MaxBufSize;
      }

   // If new buffer isn't big enough to hold existing
   // data, we must stop now, since it has already
   // been acknowledged
   if (s->sdatalen > len)
      return 0;

   // If there was any old data and pointer changed,
   // copy it to the new spot, at offset zero

   if (s->sdatalen && dp != s->sdata) {
      up_to_end = s->maxsdatalen - s->sdataoffset;
      if (up_to_end < s->sdatalen) {
         memcpy(dp, s->sdata + s->sdataoffset, up_to_end);
         memcpy(dp+up_to_end, s->sdata, s->sdatalen - up_to_end);
         }
      else
         memcpy(dp, s->sdata + s->sdataoffset, s->sdatalen);

      s->sdataoffset = 0;
      }

   // Now we can change the pointer and length
   s->maxsdatalen = len;
   s->sdata = dp;

   // Successful!
   return s->maxsdatalen;
   }

void sock_noflush(sock_type *s)
{
   if (s->tcp.ip_type == IPPROTO_TCP) {
      s->tcp.flags &= ~tcp_FlagPUSH;
      s->tcp.sock_mode |= TCP_LOCAL ;
      }
   }

void sock_flush(sock_type *s)
{
    if (s->tcp.ip_type == IPPROTO_TCP) {
        s->tcp.sock_mode &= ~TCP_LOCAL;
        tcp_Flush(&s->tcp);
    }
}

/*
 * sock_flushnext - Cause next transmission to have a flush
 */
void sock_flushnext(sock_type *s)
{
   if (s->tcp.ip_type == IPPROTO_TCP) {
      s->tcp.flags |= tcp_FlagPUSH;
      s->tcp.sock_mode &= ~TCP_LOCAL;
      }
   }

/*
 * sock_putc - Put a character
 *           - No expansion but flushes on '\n'
 *           - Returns character
 */
byte sock_putc(sock_type *s, byte c)
{
   if ((c == '\n') || (c == '\r'))
      sock_flushnext(s);
   sock_write(s, &c, 1);
   return c;
   }

byte sock_getc(sock_type *s)
{
   char ch;
   return sock_read(s, &ch, 1) < 1 ? EOF : ch;
   }

/*
 * sock_puts - Does not append carriage return in binary mode
 *           - Returns length
 */
longword sock_puts(sock_type *s, byte *dp)
{
   longword len;
   word oldmode;

   len = strlen(dp);

   if (s->tcp.sock_mode & TCP_MODE_ASCII) {
      if (s->tcp.ip_type == IPPROTO_TCP)
         s->tcp.sock_mode |= TCP_LOCAL;
      sock_noflush(s);
      if (len) sock_write(s, dp, len);
      sock_flushnext(s);
      sock_write(s, "\r\n", 2);
      }
   else {
      sock_flushnext(s);
      sock_write(s, dp, len);
      }
   return len;
   }

/*
 * sock_update - Update the socket window size to the other guy
 */
static void sock_update(tcp_Socket *s)
{
   if (s->ip_type == IPPROTO_TCP) {
      if (!s->rdatalen)
         tcp_send(s, SEND_NORMAL, "sock_update()");  /* Update the window */
      else {
#ifdef N_DEBUG_SOON
         printf("sendsoon from sock_update()\n");
#endif
         tcp_sendsoon(s);
         }
      }
   }

/*
 * sock_gets - Read a string from any socket
 *           - Return length of returned string
 *           - Removes end of line terminator(s)
 *
 *           - Quentin Smart fixed some problems
 *
 * ignores whether socket is TCP or UDP         why?
 * ignores whether socket mode is ASCII         why?
 */
longword sock_gets(sock_type *s, byte *dp, longword n)
{
   longword len, templen, up_to_end;
   char *src_p, *temp, *temp2;

   if (s->udp.ip_type != IPPROTO_UDP
         && s->udp.ip_type != IPPROTO_TCP)
      return 0;  /* Do nothing when socket is closed */

   src_p = s->udp.rdata + s->udp.rdataoffset;
   if (s->udp.rdatalen == 0) return 0;

   // Eat trailing \n or \0 from previous line
   if (*src_p == 0 || *src_p == '\n') {
      --s->udp.rdatalen;
      s->udp.rdataoffset = wrap(s->udp.rdataoffset+1, s->udp.maxrdatalen);
      if (!s->udp.rdatalen) return 0;
      }

   if (--n > s->udp.rdatalen) n = s->udp.rdatalen;

   // Copy everything
   up_to_end = s->udp.maxrdatalen - s->udp.rdataoffset;
   if (up_to_end < n) {
      memcpy(dp, s->udp.rdata + s->udp.rdataoffset, up_to_end);
      memcpy(dp+up_to_end, s->udp.rdata, n-up_to_end);
      }
   else
      memcpy(dp, s->udp.rdata + s->udp.rdataoffset, n);
   dp[n] = 0;  /* Terminate new string */

   temp = memchr(dp, '\n', n);
   temp2= memchr(dp, '\r', n);
   if (temp) *temp = 0;
   if (temp2) *temp2= 0;

   // Fail if there were no carriage returns
   if (!temp2) {
      *dp = 0;
      return 0;
      }

   // Joel 6/26/96 fixed bug: wasn't subtracting dp
   // when both \n and \r had been found
   //
   len = temp ? ptr_min(temp, temp2)-dp : temp2-dp;

   // expect \r\n or \r\0 or \n or \r
   // so get first of the two

   n = len+1;                // account for first \r or \n

   // We check next char if it exists, and skip it if 0, \r, or \n
   if (s->udp.rdatalen > n &&
         !s->udp.rdata[wrap(s->udp.rdataoffset+n, s->udp.maxrdatalen)])
      ++n;
   s->udp.rdataoffset = wrap(s->udp.rdataoffset+n, s->udp.maxrdatalen);
   if (s->udp.rdatalen > n) s->udp.rdatalen -= n;
   else s->udp.rdatalen = 0;

   sock_update(&s->tcp);   /* New window */
   return len;
   }


/*
 * sock_dataready - Returns number of bytes waiting to be read
 *                - If in ASCII mode, return 0 until a line is present
 *                  or the buffer is full
 */
longword sock_dataready(sock_type *s)
{
   longword len, up_to_end;
   char *p;

   if (!(len = s->tcp.rdatalen)) return 0;
   if (s->tcp.sock_mode & TCP_MODE_ASCII) {
      p = s->tcp.rdata + s->tcp.rdataoffset;
      if (*p == '\n') {
         s->tcp.rdataoffset = wrap(s->tcp.rdataoffset+1, s->tcp.maxrdatalen);
         s->tcp.rdatalen = --len;
         if (!len) return 0;
         p = s->tcp.rdata + s->tcp.rdataoffset;
         }
      /* Check for terminating \r */
      up_to_end = s->tcp.maxrdatalen - s->tcp.rdataoffset;
      if (up_to_end < len) {
         if (memchr(p, '\r', up_to_end))
            return len;
         if (memchr(s->tcp.rdata, '\r', len-up_to_end))
            return len;
         }
      else
         if (memchr(p, '\r', len))
            return len;
      return 0;
      }
   else
      return len;
   }

// This used to also return TRUE when TCP
// state was ESTCL or CLOSWT
//
// I couldn't find any callers, and I wanted a function
// that only returned TRUE when ready to TX, so I changed it
//
// Someday maybe create separate functions that tell you
// whether you can TX and RX
//
//              Joel Apisdorf 7/3/96
//
sock_established(sock_type *s)
{
   switch (s->tcp.ip_type) {
   case IPPROTO_UDP:
      return 1;
   case IPPROTO_TCP:
      return s->tcp.state == tcp_StateESTAB;
   default:
      return 0;
      }
   }

void sock_close(sock_type *s)
{
   switch (s->udp.ip_type) {
   case IPPROTO_UDP :
      udp_close(&s->udp);
      break;
   case IPPROTO_TCP :
      tcp_close(&s->tcp);
      tcp_tick(s);
      break;
      }
   }

_ip_delay0(sock_type *s, int timeoutseconds, waitprocref fn, int *statusptr)
/* Called by macro sock_wait_established() */
{
   int status;
   ip_timer_init(&s->udp, timeoutseconds);
   do {
      if (s->tcp.ip_type == IPPROTO_TCP) {
         if (tcp_established(&s->tcp)) {
            status = 0;
            break;
            }
         }
      kbhit();   /* Permit ^c */
      if (!tcp_tick(s)) {
         if (!s->tcp.err_msg) s->tcp.err_msg = "Host refused connection";
         status = -1;       /* get an early reset */
         break;
         }
      if (ip_timer_expired(&s->udp)) {
         s->tcp.err_msg = "Open timed out";
         sock_close(s);
         status = -1;
         break;
         }

      if (fn) {
         status = fn(s);
         if (status)
            break;
         }

      if (s->tcp.usr_yield)
         (*s->tcp.usr_yield)();

      if (s->tcp.ip_type == IPPROTO_UDP) {
         status = 0;
         break;
         }
      } while (1);
   if (statusptr) *statusptr = status;
   return status;
   }

_ip_delay1(sock_type *s, int timeoutseconds, waitprocref fn, int *statusptr)
/* Called by macro sock_wait_input() */
{
   int status;
   ip_timer_init(&s->udp, timeoutseconds);

   sock_flush(s);  /* New enhancement */

   do {
      if (sock_dataready(s)) {
         status = 0;
         break;
         }
      kbhit();  /* Permit ^c */

      if (!tcp_tick(s)) {
         status = 1;
         break;
         }
      if (ip_timer_expired(&s->udp)) {
         s->tcp.err_msg = "Connection timed out";
         sock_close(s);
         status = -1;
         break;
         }
      if (fn) {
         status = fn(s);
         if (status)
            break;
         }
      if (s->tcp.usr_yield)
         (*s->tcp.usr_yield)();
      } while (1);
   if (statusptr) *statusptr = status;
   return status;
   }

_ip_delay2(sock_type *s, int timeoutseconds, waitprocref fn, int *statusptr)
/* Called by macro sock_wait_closed() */
{
   int status;
   ip_timer_init(&s->udp, timeoutseconds);

   if (s->tcp.ip_type != IPPROTO_TCP) {
      if (statusptr) * statusptr = 1;
      return 1;
      }

   do {
      /* In this situation we KNOW user not planning to read rdata */
      s->tcp.rdatalen = 0;
      kbhit();  /* Permit ^c */
      if (!tcp_tick(s)) {
         status = 1;
         break;
         }
      if (ip_timer_expired(&s->udp)) {
         s->tcp.err_msg = "Connection timed out";
         sock_abort(s);
         status = -1;
         break;
         }
      if (fn) {
         status = fn(s);
         if (status)
            break;
         }
      if (s->tcp.usr_yield) (*s->tcp.usr_yield)();

      } while (1);
   if (statusptr) *statusptr = status;
   return status;
   }


char *rip(char *s)  // Replace \n and \r by \0 in a string
{
   char *temp;

   temp = strchr(s, '\n');
   if (temp)
      *temp = 0;

   temp = strchr(s, '\r');
   if (temp)
      *temp = 0;

   return s;
   }
